/* GPLv2 applies
 * SVN revision: $Revision: 661 $
 * (C) 2006-2013 by folkert@vanheusden.com
 */
#define _GNU_SOURCE
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ncursesw/panel.h>
#include <ncursesw/ncurses.h>

#include "gen.h"
#include "error.h"
#include "theme.h"
#include "term.h"
#include "buffer.h"
#include "channels.h"
#include "servers.h"
#include "utils.h"
#include "tcp.h"
#include "loop.h"
#include "irc.h"
#include "names.h"
#include "user.h"
#include "key_value.h"
#include "wordcloud.h"
#include "grep_filter.h"
#include "config.h"

BOOL notified_grep_filter_problem = FALSE;

BOOL is_channel(const char *name)
{
	return name[0] == '#' || name[0] == '&';
}

int do_send(int fd, char *format, ...)
{
	/* not to way to do it but I can't find a way to "predict" how large
	 * a string will be
	 */
	char *str_buffer = NULL;
	int len = 0, rc = -1;
        va_list ap;

        va_start(ap, format);
	len = vasprintf(&str_buffer, format, ap);
        va_end(ap);

	str_buffer = realloc(str_buffer, len + 3);
	str_buffer[len + 0] = '\r';
	str_buffer[len + 1] = '\n';
	str_buffer[len + 2] = 0x00;

	LOG("OUT %d: %s\n", fd, str_buffer);

	len += 2; /* CR/LF */

	rc = WRITE(fd, str_buffer, len);
	if (rc != len)
	{
		LOG("send: %d, rc: %d\n", len, rc);
		myfree(str_buffer);
		return -1;
	}

	myfree(str_buffer);

	return 0;
}

int irc_nick(int fd, const char *nick)
{
	if (do_send(fd, "NICK %s", nick))
		return -1;

	return 0;
}

int irc_login1(server *pserver)
{
	/* password required? */
	if (pserver -> password)
	{
		if (do_send(pserver -> fd, "PASS %s", pserver -> password) == -1)
			return -1;
	}

	/* send nick */
	if (irc_nick(pserver -> fd, pserver -> nickname) == -1)
		return -1;

	return 0;
}

int irc_login2(server *pserver)
{
	/* logon */
	if (do_send(pserver -> fd, "USER %s %s %s :%s", pserver -> username, "-", pserver -> server_host, pserver -> user_complete_name) == -1)
		return -1;

	return 0;
}

int create_channel(int server_index, char *channel)
{
	/* find channel */
	int channel_index = find_channel_index(server_index, channel);
	if (channel_index == -1)
	{
		channel_index = add_channel(server_index, channel);

		/* recreate list of channels */
		show_channel_names_list();
	}

	return channel_index;
}

void set_completename_if_empty(int cur_server, char *complete_name)
{
	char *dummy = strrchr(complete_name, '!');

	if (dummy)
	{
		int channel_index;

		*dummy = 0x00;

		for(channel_index = 0; channel_index < server_list[cur_server].n_channels; channel_index++)
		{
		 	channel *cur_channel = &server_list[cur_server].pchannels[channel_index];
			int name_index = find_nick(cur_server, channel_index, complete_name);

			if (name_index != -1)
			{
				if (strcmp(cur_channel -> persons[name_index].name_complete, "") == 0)
				{
					myfree(cur_channel -> persons[name_index].name_complete);
					cur_channel -> persons[name_index].name_complete = strdup(dummy + 1);
				}
			}
		}
	}
}

int process_server_do_line(int cur_server, char *string)
{
	int rc = 0;
	char *space;
	char *cur = string;
	char *pars = NULL;
	/* command splitup */
	char *pre = NULL;
	char *cmd = NULL;
	time_t now = time(NULL);

	LOG("%s %d> %s (%c)\n", server_list[cur_server].description ? server_list[cur_server].description : server_list[cur_server].server_host, server_list[cur_server].fd, string, cur[0]);

	/* get pre-thing */
	if (cur[0] == ':')
	{
		space = strchr(cur, ' ');
		if (!space)
		{
			LOG("!! line without spaces behind the :\n");
			return -1;
		}

		*space = 0x00;
		pre = strdup(cur + 1);

		space++;
		while(*space == ' ') space++;
		cur = space;
	}

	/* get command */
	/* find end of command */
	space = strchr(cur, ' ');
	if (!space)
		space = &cur[strlen(cur)];
	*space = 0x00;
	space++;

	while(*space == ' ')
		space++;

	cmd = strdup(cur);

	/* get parameter(s) */
	pars = space;

	LOG(">IN: cmd: '%s', pars: '%s', pre: '%s'\n", str_or_nothing(cmd), str_or_nothing(pars), str_or_nothing(pre));

	/* do commands */
	if (strcmp(cmd, "PING") == 0)			/* PING/PONG */
	{
		if (*pars == ':') pars++;

		if (do_send(server_list[cur_server].fd, "PONG %s", pars) == -1)
			rc = -1;
	}
	else if (strcmp(cmd, "TOPIC") == 0)
	{
		char *space = strchr(pars, ' '), *colon;
		if (space)
		{
			char *newsubj = NULL;
			int channel_index;

			*space = 0x00;
			channel_index = create_channel(cur_server, pars);

			myfree(server_list[cur_server].pchannels[channel_index].topic);

			colon = strchr(space + 1, ':');
			if (colon)
			{
				newsubj = colon + 1;
			}
			else
			{
				LOG("!! topic missing : -> continuing\n");
				newsubj = space + 1;
			}

			server_list[cur_server].pchannels[channel_index].topic = strdup(newsubj);

			if (server_list[cur_server].pchannels[channel_index].keeptopic != NULL &&
			    strstr(newsubj, server_list[cur_server].pchannels[channel_index].keeptopic) == NULL)
			{
				if (irc_topic(server_list[cur_server].fd,
					      server_list[cur_server].pchannels[channel_index].channel_name,
					      server_list[cur_server].pchannels[channel_index].keeptopic) == -1)
					rc = -1;
			}

			server_list[cur_server].pchannels[channel_index].topic_changed = TRUE;

			update_statusline(cur_server, channel_index, "topic changed on channel %s by %s to %s", server_list[cur_server].pchannels[channel_index].channel_name, pre, space + 1);

			if (pre)
				set_completename_if_empty(cur_server, pre);
		}
		else
		{
			LOG("!! topic command without space\n");
		}
	}
	else if (strcmp(cmd, "KICK") == 0)
	{
		int channel_index;
		/* KICK #knageroe flok :reason */
		char *colon = strchr(pars, ':');
		/* pars points to the channel */
		char *kick_nick = NULL, *dummy = strchr(pars, ' ');
		if (dummy)
		{
			*dummy = 0x00;
			kick_nick = dummy + 1;
			dummy = strchr(kick_nick, ' ');
			if (dummy) *dummy = 0x00;
		}

		channel_index = create_channel(cur_server, pars);
		if (channel_index != -1 && kick_nick != NULL)
		{
			char redraw = 0;

			if (colon)
				colon++;

			if (cur_server == current_server && current_server_channel_nr == channel_index)
				redraw = 1;

			/* the user that kicks is given in the pre-stuff, the user
			 * that gets kicked is in the pars
			 */
			/* me? */
			if (strcasecmp(kick_nick, server_list[cur_server].nickname) == 0)
			{
				update_statusline(cur_server, 0, "you were kicked from channel '%s' by %s: %s\n", pars, str_or_nothing(pre), str_or_nothing(colon));
				cmd_LEAVE(cur_server, channel_index, pars);

				if (auto_rejoin)
					irc_join(server_list[cur_server].fd, pars);
				else
					popup_notify(FALSE, "You were kicked from channel %s by\n%s, reason:\n%s", pars, str_or_nothing(pre), str_or_nothing(colon));
			}
			else
			{
				delete_name(cur_server, channel_index, kick_nick, NULL);
				update_statusline(cur_server, channel_index, "user '%s' kicked from channel '%s' by %s\n", str_or_nothing(kick_nick), pars, str_or_nothing(pre));
			}

			if (redraw)
			{
				show_channel_names_list();
			}
		}
		else if (channel_index == -1)
			update_statusline(cur_server, 0, "user '%s' kicked from unknown channel '%s'\n", str_or_nothing(kick_nick), pars);
		else
		{
			LOG("!! no nick for kick!\n");
			update_statusline(cur_server, channel_index, "no nick for kick?!\n");
		}

		if (pre)
			set_completename_if_empty(cur_server, pre);
	}
	else if (strcmp(cmd, "PART") == 0 && pre != NULL && pars != NULL)	/* PART (leave channel) */
	{
		int channel_index;
		/* pars points to the channel */
		char *dummy = strchr(pars, ' ');
		if (dummy) *dummy = 0x00;

		/* for some reason, sometimes there's a ':' before the channel name */
		if (pars[0] == ':')
			pars++;

		channel_index = find_channel_index(cur_server, pars);
		if (channel_index != -1)
		{
			/* the user is given in the pre-stuff */
			delete_name(cur_server, channel_index, NULL, pre);

			if (show_parts)
				update_statusline(cur_server, channel_index, "user %s left %s", pre, pars);
		}

		if (pre)
			set_completename_if_empty(cur_server, pre);
	}
	else if (strcmp(cmd, "MODE") == 0)		/* MODE */
	{
		char *space_after_channel = strchr(pars, ' ');
		if (space_after_channel && pre)
		{
			/* MODE (#)channel mode (nick) */
			/*         pars    mode  nick  */
			int channel_index = 0;
			char *mode = space_after_channel + 1;
			char *dummy = NULL;
			char *space_after_mode = NULL;
			*space_after_channel = 0x00;

			while(*mode == ' ')
				mode++;

			space_after_mode = strchr(mode, ' ');
			if (space_after_mode)
				*space_after_mode = 0x00;

			dummy = strrchr(pre, '!');
			if (dummy)
				*dummy = 0x00;

			if (is_channel(pars) && space_after_mode != NULL)
			{
				char *nick = space_after_mode + 1;

				while(*nick == ' ')
					nick++;

				dummy = strchr(nick, ' ');
				if (dummy)
					*dummy = 0x00;

				/* find channel */
				channel_index = create_channel(cur_server, pars);
				if (channel_index != -1)
				{
					char ch_mode = 0;

					if (mode[1] == 'o')
						ch_mode = MODE_OPERATOR;
					else if (mode[1] == '+')
						ch_mode = MODE_CAN_SPEAK;

					if (mode[0] == '+')
						set_name_mode(cur_server, channel_index, nick, get_nick_mode(nick) | ch_mode);
					else if (mode[0] == '-')
						set_name_mode(cur_server, channel_index, nick, get_nick_mode(nick) & (~ch_mode));
					else
					{
						LOG("!! unknown mode thing '%s' in channel '%s'\n", mode, pars);
					}

					show_channel_names_list();
				}
				else
				{
					LOG("!! failed to find/create channel '%s'\n", pars);
					channel_index = 0;
				}

				if (show_mode_changes)
					update_statusline(cur_server, channel_index, "new mode %s by %s for %s in channel %s", mode, pre, nick, pars);
			}
			else if (show_mode_changes)
			{
				update_statusline(cur_server, channel_index, "new mode %s for %s", str_or_nothing(mode), str_or_nothing(pars));
			}
		}
		else
		{
			LOG("!! no mode thing\n");
		}

		if (pre) set_completename_if_empty(cur_server, pre);
	}
	else if (strcmp(cmd, "QUIT") == 0)		/* QUIT */
	{
		/* delete from all names-lists for this server */
		if (pre && strlen(pre) > 0)
		{
			char *colon = strchr(pars, ':'), *msg = colon ? colon + 1 : "";
			server *ps = &server_list[cur_server];
			int loop;

			for(loop=0; loop<ps -> n_channels; loop++)
			{
				/* the user is given in the pre-stuff */
				if (delete_name(cur_server, loop, NULL, pre))
				{
					if (show_joins)
						update_statusline(cur_server, loop, "user %s left %s: %s", pre, ps -> pchannels[loop].channel_name, msg);
				}
			}
		}
		else
		{
			LOG("!! no user for 'QUIT'");
		}
	}
	else if (strcmp(cmd, "NICK") == 0)		/* NICK */
	{
		/* server acknowledges (new) NICK name */
		char *colon = strchr(pars, ':'); /* nick is behind a colon */

		if (colon == NULL) /* no colon? then = parameter */
			colon = pars;
		else
			colon++;

		if (colon != NULL && pre != NULL)
		{
			char *current_nick = strdup(pre);
			char *exclamation_mark = strrchr(current_nick, '!');

			if (exclamation_mark)
				*exclamation_mark = 0x00;

			if (strcasecmp(current_nick, server_list[cur_server].nickname) == 0)/* acknowledgement to my a nick */
			{							/* change I did myself? */
				myfree(server_list[cur_server].nickname);
				server_list[cur_server].nickname = strdup(colon);
			}
			else
			{
				int channel_index;

				for(channel_index = 0; channel_index < server_list[cur_server].n_channels; channel_index++)
				{
					/* change nick in each channel */
					if (find_nick(cur_server, channel_index, current_nick) != -1)
					{
						change_nick(cur_server, channel_index, colon, pre);

						if (show_nick_change)
							update_statusline(cur_server, channel_index, "%s is now known as %s", pre, colon);
					}

					/* change channelnames */
					if (strcasecmp(server_list[cur_server].pchannels[channel_index].channel_name, current_nick) == 0)
					{
						myfree(server_list[cur_server].pchannels[channel_index].channel_name);
						server_list[cur_server].pchannels[channel_index].channel_name = strdup(colon);
					}
				}
			}

			myfree(current_nick);
		}
		else
		{
			LOG("!! invalid NICK command by server\n");
		}
	}
	else if (strcmp(cmd, "JOIN") == 0)		/* JOIN */
	{ 	/* server acknowledges channel join */
		char *colon = strchr(pars, ':'); /* channel is behind a colon */
		if (!colon)	/* work around bugs in servers */
		{
			LOG("!! no colon behind JOIN\n");
			colon = pars;
		}
		else
			colon++; /* ignore the colon */

		if (colon)
		{
			int channel_index = -1;
			char *dummy = NULL, *dummy2 = NULL;

			/* skip over spaces */
			while(*colon == ' ')
				colon++;

			/* if there's anything behind the channel, cut it off here */
			dummy = strchr(colon, ' ');
			if (dummy)
				*dummy = 0x00;

			/* only add when we don't have this channel (in case the
			 * server screwed up)
			 */
			channel_index = create_channel(cur_server, colon);

			/* add user to names list */
			if (channel_index != -1)
			{
				if (pre && strlen(pre) > 0)
				{
					if (find_name(cur_server, channel_index, pre) == -1)
						add_name(cur_server, channel_index, NULL, pre);
					else
						change_nick(cur_server, channel_index, NULL, pre);

					if (show_joins)
						update_statusline(cur_server, channel_index, "user %s joined %s", pre, colon);
				}
				else
				{
					LOG("!! no user for JOIN\n");
				}
			}
			else
			{
				LOG("!! JOIN: channel %s could not be created\n", colon);
			}

			/* myself? */
			dummy = strdup(pre);
			dummy2 = strrchr(dummy, '!');
			if (dummy2) *dummy2 = 0x00;
			if (strcasecmp(server_list[cur_server].nickname, dummy) == 0)
			{
				/* get a list of hostnames (using the WHO command) */
				if (irc_who(server_list[cur_server].fd, server_list[cur_server].pchannels[channel_index].channel_name) == -1)
					rc = -1;
			}

			myfree(dummy);

			show_channel_names_list();
		}
		else
		{
			LOG("!! server acknowledges JOIN but doesn't return channelname\n");
		}
	}
	else if (strcmp(cmd, "INVITE") == 0)	/* INVITE */
	{
		char *colon = strchr(pars, ':');

		if (!colon)
			update_statusline(cur_server, 0, "Received garbled INVITE message (channel missing)");
		else if (!allow_invite)
			update_statusline(cur_server, 0, "Ignoring invite for channel %s", colon + 1);
		else
			rc = irc_join(server_list[cur_server].fd, colon + 1);
	}
	else if (strcmp(cmd, "PRIVMSG") == 0 ||
		 strcmp(cmd, "NOTICE") == 0)	/* PRIVMSG / NOTICE / set topic */
	{
		char *colon = strchr(pars, ':');
		int channel_index = 0;
		server *ps = &server_list[cur_server];

		if (ps -> ts_last_msg)
		{
			ps -> t_event += now - ps -> ts_last_msg;
			ps -> n_event++;

			if (ps -> n_event > 100)
			{
				ps -> t_event /= (double)ps -> n_event;

				ps -> t_event *= 5;
				ps -> n_event = 5;
			}
		}

		ps -> ts_last_msg = now;

		if (notice_in_server_channel == TRUE && strcmp(cmd, "NOTICE") == 0)
			channel_index = 0;
		else
		{
			char *ch = NULL, *space = NULL;
			char *err = NULL;

			/* first thing behind "PRIVMSG" is the channel */
			while(*pars == ' ')
				pars++;

			space = strchr(pars, ' ');
			if (space)
				*space = 0x00;
			else
				space = "";

			while(*space == ' ')
				space++;

			/* is this a private msg? (dest is my nick: use * nick of sender) */
			if (strcasecmp(server_list[cur_server].nickname, pars) == 0)
			{
				char *dummy = NULL;

				if (pre)
					ch = strdup(pre);
				else
					ch = strdup("(unknown)");

				dummy = strrchr(ch, '!');

				if (dummy)
					*dummy = 0x00;
			}
			else
			{
				ch = strdup(pars);
			}

			if (colon && process_grep_filter(ps -> description, ch, colon + 1, &err) == FALSE && notified_grep_filter_problem == FALSE)
			{
				popup_notify(FALSE, "%s", err);
				free(err);

				notified_grep_filter_problem = TRUE;
			}

			/* find channel index */
			channel_index = create_channel(cur_server, ch);

			myfree(ch);
		}

		if (colon)
		{
			const char *msg_text = colon + 1;

			add_to_wc(msg_text);

			if (log_channel(cur_server, channel_index, pre, msg_text, FALSE) == -1)
				rc = -1;
		}
		else
		{
			LOG("!! no text for PRIVMSG/NOTICE\n");
		}

		if (strcmp(cmd, "PRIVMSG") == 0)
			ps -> pchannels[channel_index].recvd_non_notice = TRUE;

		/* in case F-IRC gets really slow, remove this line: */
		if (pre)
			set_completename_if_empty(cur_server, pre);
	}
	else if (cmd[0] >= '0' && cmd[0] <= '9')	/* error? */
	{
		char *colon = strchr(pars, ':');
		int err = atoi(cmd);
		int lc_rc;

		if (pre) set_completename_if_empty(cur_server, pre);

		if (colon)
		{
			colon++;

			while(*colon == ' ') colon++;
		}

		LOG("got replycode: %d (%s)\n", err, pars);

		if (err == 001)
		{
			update_statusline(cur_server, 0, "Logged in to %s:%d (%s)", server_list[cur_server].server_host, (int)server_list[cur_server].server_port, str_or_nothing(server_list[cur_server].description));

			set_state(cur_server, STATE_RUNNING);
		}
		else if (err == 317)		/* idle time and such */
		{
			time_t signon = 0;

			/* flok Uhmmm 111 1110956321 :seconds idle, signon time */
			char *space = strchr(pars, ' ');
			if (space)
			{
				char *nick = NULL;

				while(*space == ' ')
					space++;

				/* space now points behind 'flok' */
				nick = space;
				space = strchr(space, ' ');
				if (space)
				{
					int idle = 0;

					*space = 0x00;
					space++;

					 while(*space == ' ')
						space++;

					/* space now points behind 'uhmmm' */
					idle = atoi(space);

					space = strchr(space, ' ');
					if (space)
					{
						char *str_buffer = NULL;

						while(*space == ' ')
							space++;

						/* space now points behind 111 */
						signon = (time_t)atol(space);

						asprintf(&str_buffer, "%s is %d seconds idle and on-line since %s", nick, idle, ctime(&signon));

						lc_rc = log_channel(cur_server, 0, NULL, str_buffer, TRUE);
						myfree(str_buffer);

						if (lc_rc == -1)
							rc = -1;
					}
				}
			}
		}
		else if (err == 332)		/* set topic */
		{
			/* find channel index */
			char *dummy = pars;
			char *ch = NULL;

			/* find channel */
			while(*dummy == ' ')
				dummy++;

			ch = strchr(dummy, ' ');
			if (!ch)
				update_statusline(cur_server, 0, "invalid 332 line from server");
			else
			{
				int channel_index = 0;

				while(*ch == ' ')
					ch++;

				dummy = strchr(ch, ' ');
				if (dummy)
					*dummy = 0x00;

				channel_index = create_channel(cur_server, ch);

				myfree(server_list[cur_server].pchannels[channel_index].topic);

				if (colon)
					server_list[cur_server].pchannels[channel_index].topic = strdup(colon);
				else
					server_list[cur_server].pchannels[channel_index].topic = NULL;


				server_list[cur_server].pchannels[channel_index].topic_changed = TRUE;
			}
		}
		else if (err == 409)		/* ":No origin specified": PING/PONG fails! */
		{
			update_statusline(cur_server, 0, "internal error while replying to PING");
			rc = -1;
		}
		else if (err >= 431 && err <= 436)/* problem setting NICK */
		{
			if (server_list[cur_server].state != STATE_RUNNING)
			{
				char *msg = NULL;
				const char *nick = gen_random_nick();

				rc = irc_nick(server_list[cur_server].fd, nick);

				asprintf(&msg, "Your default nick was refused, using a temporary random one: %s", nick);
				update_statusline(cur_server, 0, msg);

				myfree(nick);

				myfree(msg);
			}
			else
			{
				const char msg [] = "Your new nick was refused";

				update_statusline(cur_server, 0, msg);

				popup_notify(FALSE, msg);
			}
		}
		else if (err == 353 || err == 366)	/* channels nick list */
		{
			char *channel_start = strchr(pars, '#');

			if (channel_start == NULL)
				channel_start = strchr(pars, '&');

			if (channel_start && colon)
			{
				int channel_index = 0;
				char *pnames = colon;
				char *channel_space = strchr(channel_start, ' ');
				channel *pc = NULL;

				if (channel_space)
					*channel_space = 0x00;

				/* find channel index */
				channel_index = create_channel(cur_server, channel_start);

				pc = &server_list[cur_server].pchannels[channel_index];

				if (err == 353)
				{
					if (pc -> adding_names == FALSE)
					{
						free_names_list(pc);

						pc -> adding_names = TRUE;
					}

					for(; strlen(pnames) > 0;)
					{
						char *space = strchr(pnames, ' ');
						if (space)
							*space = 0x00;

						add_name(cur_server, channel_index, pnames, NULL);

						if (!space)
							break;

						pnames = space + 1;
					}
				}
				else
				{
					pc -> adding_names = FALSE;
				}

				show_channel_names_list();
			}
			else
			{
				LOG("!! empty channellist\n");
			}
		}
		else if (err == 311)	/* reply to whois */
		{
			char *dummy = NULL, *hisnick = NULL;
			char *colon = strchr(pars, ':');

		        /* 311     RPL_WHOISUSER
                           "<nick> <user> <host> * :<real name>" */
			/* :irc.xs4all.nl 311 mynick hisnick hisuser hishost * :hisname */

			/* points to after mynick */
			dummy = strchr(pars, ' ');
			while(dummy != NULL && isspace(*dummy))
				dummy++;
			hisnick = dummy;

			if (hisnick)
			{
				/* points after hisuser */
				dummy = strchr(hisnick, ' ');
				if (dummy)
				{
					char str_buffer[IRC_MAX_MSG_LEN + 1] = { 0 };
					char *hisuser = NULL;
					int channel_index = -1;

					*dummy = 0x00;
					hisuser = dummy + 1;

					dummy = strchr(hisuser, ' ');
					if (dummy)
						*dummy = '@';

					dummy = strchr(hisuser, ' ');
					if (dummy)
						*dummy = 0x00;

					snprintf(str_buffer, sizeof str_buffer, "%s (%s) is %s", hisnick, hisuser, str_or_nothing(colon + 1));
					lc_rc = log_channel(cur_server, 0, NULL, str_buffer, TRUE);
					if (lc_rc == -1)
						rc = -1;

					snprintf(str_buffer, sizeof str_buffer, "%s!%s", hisnick, hisuser);
					str_buffer[sizeof(str_buffer) - 1]=0x00;

					/* change nick in each channel */
					for(channel_index = 0; channel_index < server_list[cur_server].n_channels; channel_index++)
					{
						if (find_nick(cur_server, channel_index, hisnick) != -1)
							change_nick(cur_server, channel_index, hisnick, str_buffer);
					}
				}
			}
		}
		else if (err == 391)		/* TIME */
		{
			char *colon = strchr(pars, ':');
			server *ps = &server_list[cur_server];

			if (!ps -> sent_time_req && colon && (ps -> prev_cmd == NULL || strcmp(ps -> prev_cmd, cmd) != 0))
				lc_rc = log_channel(cur_server, 0, NULL, colon + 1, TRUE);

			ps -> sent_time_req = FALSE;
		}
		else if (err == 321)		/* LIST start */
		{
			int index = 0;
			server *ps = &server_list[cur_server];

			for(index=0; index<ps -> n_all_channels; index++)
				myfree(ps -> all_channels[index]);

			myfree(ps -> all_channels);

			ps -> n_all_channels = 0;
			ps -> all_channels = NULL;

			ps -> all_channels_complete = FALSE;
		}
		else if (err == 322)		/* LIST part */
		{
			char *space = strchr(pars, ' '); /* find end of some nick, start of channel */

			if (space)
			{
				char *ch_start = space + 1;

				space = strchr(ch_start, ' ');
				if (space)
				{
					server *p = &server_list[cur_server];

					*space = 0x00;

					p -> all_channels = (const char **)realloc(p -> all_channels, sizeof(const char *) * (p -> n_all_channels + 1));
					p -> all_channels[p -> n_all_channels] = strdup(ch_start);
					p -> n_all_channels++;
				}
			}
		}
		else if (err == 323)		/* LIST end */
		{
			server_list[cur_server].all_channels_complete = TRUE;
		}
		else if (err == 352)		/* reply to 'WHO' */
		{
			/* <usernick> <channel> <user> <host> <server> <nick> */
			for(;;)
			{
				int nick_index = -1;
				channel *cur_channel = NULL;
				char *dummy = strchr(pars, ' '); /* points before channel */
				int channel_index = 0;
				char *dummy2 = NULL, *dummy3 = NULL;

				if (!dummy)
				{
					LOG("!! 352 malformed (start of channel)\n");
					break;
				}

				/* find start of channel */
				while(isspace(*dummy))
					dummy++;
				dummy2 = strchr(dummy, ' '); /* end of channel, start of user */
				if (!dummy2)
				{
					LOG("!! 352 malformed (end of channel)\n");
					break;
				}
				*dummy2 = 0x00;

				channel_index = create_channel(cur_server, dummy);
				if (channel_index == -1)
				{
					LOG("!! 352 could not find/create channel '%s'\n", dummy);
					break;
				}

				/* find start of user */
				dummy = dummy2 + 1;
				while(isspace(*dummy))
					dummy++;
				/* find space between user and host */
				dummy2 = strchr(dummy, ' ');
				if (!dummy2)
				{
					LOG("!! 352 host missing\n");
					break;
				}
				*dummy2 = '@';

				/* find end of user@host */
				dummy2 = strchr(dummy2, ' ');
				if (!dummy2)
				{
					LOG("!! 352 server missing\n");
					break;
				}
				*dummy2 = 0x00;
				/* dummy now points to user@host */

				/* find start of server */
				dummy2++;
				while(isspace(*dummy2))
					dummy2++;
				/* find end of server */
				dummy2 = strchr(dummy2, ' ');
				if (!dummy2)
				{
					LOG("!! 352 server is ending\n");
					break;
				}

				/* find start of nick */
				while(isspace(*dummy2))
					dummy2++;

				dummy3 = strchr(dummy2, ' ');
				if (dummy3)
					*dummy3 = 0x00;
				/* dummy2 now points to the start of the nick*/

				nick_index = find_nick(cur_server, channel_index, dummy2);
				if (nick_index == -1)
				{
					LOG("!! 352 nick %s not found!\n", dummy2);
					break;
				}

				cur_channel = &server_list[cur_server].pchannels[channel_index];

				myfree((cur_channel -> persons)[nick_index].name_complete);
				(cur_channel -> persons)[nick_index].name_complete = strdup(dummy);

				break;
			}
		}
		else
		{
			int nlen = strlen(server_list[cur_server].nickname);
			const char *p = str_or_nothing(pre), *ppars = pars;
			char str_buffer[IRC_MAX_MSG_LEN + 1] = { 0 };
			int sb_len = sizeof(str_buffer) - 1;

			LOG("!! unknown error %s\n", cmd);

			if (strncmp(pars, server_list[cur_server].nickname, nlen) == 0)
			{
				ppars = strchr(pars, ' ');

				if (ppars)
					ppars++;
				else
					ppars = pars;
			}

			if (strchr(p, '@') != NULL || strchr(p, '#') != NULL)
			{
				if (colon)
					snprintf(str_buffer, sb_len, "%s: %s", str_or_nothing(pre), colon);
				else
					snprintf(str_buffer, sb_len, "%s: %s", str_or_nothing(pre), ppars);
			}
			else
			{
				if (colon)
					strncpy(str_buffer, colon, sb_len);
				else
					strncpy(str_buffer, ppars, sb_len);
			}

			str_buffer[sb_len - 1] = 0x00;

			lc_rc = log_channel(cur_server, 0, NULL, str_buffer, TRUE);

			if (lc_rc == -1)
				rc = -1;
		}
	}
	else
	{
		update_statusline(cur_server, 0, "unknown command from irc-server: ':%s %s %s'", str_or_nothing(pre), str_or_nothing(cmd), str_or_nothing(pars));
	}

	if (strcmp(cmd, "PRIVMSG") != 0 && strcmp(cmd, "NOTICE") != 0 && strcmp(cmd, "PING") != 0)
	{
		free(server_list[cur_server].prev_cmd);
		server_list[cur_server].prev_cmd = strdup(cmd);
	}

	myfree(cmd);
	myfree(pre);

	return rc;
}

int irc_privmsg(int fd, const char *channel, const char *msg)
{
	return do_send(fd, "PRIVMSG %s :%s", channel, msg);
}

int irc_kick(int fd, const char *channel, const char *nick, const char *comment)
{
	if (comment)
		return do_send(fd, "KICK %s %s :%s", channel, nick, comment);
	else
		return do_send(fd, "KICK %s %s", channel, nick);
}

int irc_ban(int fd, const char *channel, const char *nick)
{
	return do_send(fd, "MODE %s +b %s", channel, nick);
}

int irc_list(int fd)
{
	return do_send(fd, "LIST");
}

int irc_op(int fd, const char *channel, const char *nick, BOOL op)
{
	if (op)
		return do_send(fd, "MODE %s +o %s", channel, nick);
	else
		return do_send(fd, "MODE %s -o %s", channel, nick);
}

int irc_whois(int fd, const char *nick)
{
	return do_send(fd, "WHOIS %s", nick);
}

int irc_allowspeak(int fd, const char *channel, const char *nick, BOOL speak)
{
	if (speak)
		return do_send(fd, "MODE %s +v %s", channel, nick);
	else
		return do_send(fd, "MODE %s -v %s", channel, nick);
}

int irc_v3_cap_start(int fd)
{
	return do_send(fd, "CAP REQ");
}

int irc_v3_cap_end(int fd)
{
	return do_send(fd, "CAP END");
}

int irc_who(int fd, const char *channel)
{
	return do_send(fd, "WHO %s", channel);
}

int irc_topic(int fd, const char *channel, const char *topic)
{
	return do_send(fd, "TOPIC %s :%s", channel, topic);
}

int irc_ping(int fd, const char *nick, int ts)
{
	char str_buffer[IRC_MAX_MSG_LEN] = { 0 };

	snprintf(str_buffer, sizeof str_buffer, "\001PING %d\001", ts);

	return irc_privmsg(fd, nick, str_buffer);
}

int irc_version(int fd)
{
	return do_send(fd, "VERSION");
}

int irc_time(int fd)
{
	return do_send(fd, "TIME");
}

int irc_join(int fd, const char *channel)
{
	return do_send(fd, "JOIN %s", channel);
}

int irc_part(int fd, const char *channel, const char *msg)
{
	if (msg)
		return do_send(fd, "PART %s :%s", channel, msg);

	return do_send(fd, "PART %s", channel);
}

int irc_quit(int fd, const char *msg)
{
	if (msg)
		return do_send(fd, "QUIT :%s", msg);

	return do_send(fd, "QUIT");
}
