/* GPLv2 applies
 * SVN revision: $Revision: 657 $
 * (C) 2006-2013 by folkert@vanheusden.com
 */
#define _GNU_SOURCE
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ncursesw/panel.h>
#include <ncursesw/ncurses.h>

#include "gen.h"
#include "error.h"
#include "theme.h"
#include "term.h"
#include "buffer.h"
#include "channels.h"
#include "servers.h"
#include "utils.h"
#include "irc.h"
#include "loop.h"
#include "main.h"
#include "dcc.h"
#include "names.h"
#include "config.h"
#include "user.h"
#include "colors.h"
#include "ignores.h"

void popup_notify(BOOL use_getch, const char *format, ...)
{
	va_list ap;
	char *msg = NULL;
	int width = -1, loop = 0;
	string_array_t lines;
	NEWWIN *popup = NULL;

	init_string_array(&lines);

	va_start(ap, format);
	vasprintf(&msg, format, ap);
	va_end(ap);

	split_string(msg, "\n", FALSE, &lines);

	for(loop=0; loop<string_array_get_n(&lines); loop++)
	{
		int cur_width = strlen(string_array_get(&lines, loop));

		if (cur_width > width)
			width = cur_width;
	}

	popup = create_window(5 + string_array_get_n(&lines) - 1, width + 4);

	box(popup -> win, 0, 0);
	mywattron(popup -> win, A_BOLD);
	mvwprintw(popup -> win, 0, 2, "Press any key...");
	mywattroff(popup -> win, A_BOLD);

	escape_print_xy(popup, 2, 2, msg);

	mydoupdate();

	myfree(msg);

	if (use_getch)
		getch();
	else
		(void)wait_for_keypress(FALSE);

	delete_window(popup);

	free_string_array(&lines);

	mydoupdate();
}

BOOL onoff_box(const char *q, BOOL default_value)
{
	NEWWIN *bwin = NULL, *win = NULL;
	const char *title = " Space to toggle, tab select button, enter to exit ";
	int q_len = max(strlen(title), strlen(q));
	BOOL cur_value = default_value;
	BOOL ret_ok = TRUE;

	create_win_border(q_len + 5, 5, title, &bwin, &win);

	for(;;)
	{
		int c = -1;

		werase(win -> win);

		escape_print_xy(win, 1, 1, q);

		if (cur_value)
			escape_print_xy(win, 2, q_len / 2, "^ ON ^");
		else
			escape_print_xy(win, 2, q_len / 2, " OFF");

		if (ret_ok == TRUE)
			escape_print_xy(win, 3, 1, "^[ OK ]^ [ CANCEL ]");
		else
			escape_print_xy(win, 3, 1, "[ OK ] ^[ CANCEL ]^");

		mydoupdate();

		c = wait_for_keypress(FALSE);

		if (c == ' ')
			cur_value = !cur_value;
		else if (c == 9)
			ret_ok = !ret_ok;
		else if (c == 13)
		{
			if (!ret_ok)
				cur_value = default_value;

			break;
		}
		else if (c == 3)
			exit_fi();
		else if (toupper(c) == 'q')
			break;
		else
			wrong_key();
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	return cur_value;
}

const char *edit_box(int width, const char *title, const char *initial)
{
	char *line = NULL;
	NEWWIN *bwin = NULL, *win = NULL;
	BOOL ret_ok = TRUE;
	int offset = 0, pos = 0;
	BOOL reposition = TRUE;
	char *undo_clear = NULL;

	width = min(max(width, 38), max_x);

	if (initial)
		line = strdup(initial);
	else
	{
		line = (char *)malloc(1);
		line[0] = 0x00;
	}

	create_win_border(width, 3, " TAB select button, enter to exit ", &bwin, &win);

	for(;;)
	{
		char *dummy = NULL;
		int c = -1, l = strlen(line);

		werase(win -> win);

		escape_print_xy(win, 0, 1, title);

		if (ret_ok == TRUE)
			escape_print_xy(win, 2, 1, "^[ OK ]^ [ CANCEL ]");
		else
			escape_print_xy(win, 2, 1, "[ OK ] ^[ CANCEL ]^");

		if (reposition)
		{
			if (l < width)
				pos = l;
			else
				pos = (width * 3) / 4;

			offset = l - pos;

			reposition = FALSE;
		}

		dummy = strdup(&line[offset]);
		if (strlen(dummy) >= width - 2)
			dummy[width - 2] = 0x00;
		mvwprintw(win -> win, 1, 1, "%s", dummy);
		free(dummy);

		override_cursor_win = win;
		override_cursor_x = pos + 1;
		override_cursor_y = 1;

		mydoupdate();

		c = wait_for_keypress(FALSE);

		if (c == KEY_BACKSPACE || c == 127)	/* backspace/del */
		{
			if (pos > 0)
				pos--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();

			line[pos + offset] = 0x00;
		}
		else if (c == 1)	/* ^A */
			pos = offset = 0;
		else if (c == 4)	/* ^D */
		{
			int x = pos + offset;

			if (x < l)
				memmove(&line[x], &line[x + 1], l - x);
			else
				wrong_key();
		}
		else if (c == 5)	/* ^E */
			reposition = TRUE;
		else if (c == 9)
			ret_ok = !ret_ok;
		else if (c == 3)
			exit_fi();
		else if (c == 13)	/* enter */
			break;
		else if (c == 21)	/* ^U */
		{
			if (l == 0)
			{
				free(line);
				line = undo_clear;
				undo_clear = NULL;
			}
			else
			{
				free(undo_clear);
				undo_clear = strdup(line);

				line[0] = 0x00;
			}

			reposition = TRUE;
		}
		else if (c == KEY_LEFT)
		{
			if (pos > 0)
				pos--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();
		}
		else if (c == KEY_RIGHT)
		{
			int l = strlen(line);

			if (pos + offset < l)
			{
				if (pos < width - 1)
					pos++;
				else
					offset++;
			}
			else
			{
				wrong_key();
			}
		}
		else if (c == 23)	/* ^W */
		{
			int x = pos + offset;
			while(x > 0 && isalnum(line[x - 1]))
			{
				line[x - 1] = 0x00;
				x--;
			}
		}
		else if (c >= 32 && c < 127)
		{
			int l = strlen(line);
			int x = pos + offset;

			line = realloc(line, l + 1 + 1);

			if (x < l)
				memmove(&line[x + 1], &line[x], (l - x) + 1);

			line[x] = c;
			line[l + 1] = 0x00;

			if (pos < width - 1)
				pos++;
			else
				offset++;
		}
	}

	delete_window(win);
	delete_window(bwin);

	override_cursor_win = NULL;
	override_cursor_y = override_cursor_x = 0;

	mydoupdate();

	if (!ret_ok)
	{
		myfree(line);
		line = NULL;
	}

	return line;
}

void cmd_LEAVE(int server_nr, int server_channel_nr, const char *channel_name)
{
	int channel_index = -1;

	/* see if there's a channel given */
	if (channel_name && strlen(channel_name) > 0)
	{
		int loop;

		for(loop=0; loop<server_list[server_nr].n_channels; loop++)
		{
			if (strcasecmp(server_list[server_nr].pchannels[loop].channel_name, channel_name) == 0)
			{
				channel_index = loop;
				break;
			}
		}

		if (channel_index == -1)
		{
			update_statusline(server_nr, server_channel_nr, "Not in channel %s", channel_name);
			wrong_key();
		}
	}
	else
	{
		/* server channel? */
		if (server_channel_nr == 0)
		{
			update_statusline(server_nr, server_channel_nr, "Cannot leave the server channel: use /QUIT to leave a server");
			wrong_key();
		}
		else
		{
			channel_index = server_channel_nr;
		}
	}

	if (channel_index != -1)
	{
		int cur_channel = server_channel_nr;

		update_statusline(server_nr, 0, "Channel %s closed", server_list[server_nr].pchannels[channel_index].channel_name);

		close_channel(server_nr, channel_index, TRUE);

		if (server_nr == current_server)
		{
			int dummy = 0;

			if (cur_channel >= server_list[server_nr].n_channels)
				cur_channel = server_list[server_nr].n_channels - 1;

			if (cur_channel < 0)
				cur_channel = 0;

			change_channel(server_nr, cur_channel, TRUE, FALSE);

			dummy = find_vc_list_entry(server_nr, cur_channel);
			if (dummy == -1)
				dummy = 0;

                        channel_cursor = dummy % channel_window -> nlines;
                        channel_offset = dummy - channel_cursor;

			if (get_cursor_mode() == CM_NAMES)
				set_cursor_mode(CM_CHANNELS);
		}
	}

	show_channel_names_list();
}

int user_command(int current_server, int current_server_channel_nr, const char *user_line, BOOL do_command)
{
	int rc = 0;
	char me = strncasecmp(user_line, "/ME ", 4) == 0;

	if (user_line[0] == '/' && !me && do_command == TRUE) /* command for server? */
	{
		char *command = strdup(user_line + 1);
		char *pars = strchr(command, ' ');
		char *parsc = NULL;

		/* find start of parameters */
		if (pars)
		{
			*pars = 0x00;
			pars++;

			while(*pars == ' ')
				pars++;

			parsc = strdup(pars);
		}

		/* remove spaces at the end of the parameters */
		if (pars)
		{
			int pars_len = strlen(pars);

			while(pars_len)
			{
				pars_len--;

				if (isspace(pars[pars_len]))
					pars[pars_len] = 0x00;
				else
					break;
			}

			if (strlen(pars) == 0)
				pars = NULL;
		}

		/* first determine if it is a command for the irc-client or for the server */
		if (strcasecmp(command, "ADDSERVER") == 0)
		{
			if (!pars)
				popup_notify(FALSE, "/ADDSERVER: missing host(-name) to connect to");
			else
			{
				int server_index = add_server(pars,			/* host[:port] */
						server_list[current_server].username,	/* username */
						NULL,					/* password */
						server_list[current_server].nickname,	/* nickname */
						server_list[current_server].user_complete_name,	/* complete name */
						NULL					/* description */
					  );

				update_statusline(current_server, current_server_channel_nr, "Added server %s:%d", server_list[server_index].server_host, server_list[server_index].server_port);

				change_channel(server_index, 0, TRUE, TRUE);
			}
		}
		else if (strcasecmp(command, "PING") == 0)
		{
			char do_it = 1;

			if (server_list[current_server].user_ping)
			{
				if (get_ts() - server_list[current_server].t_user_ping > 10.0)
				{
					update_statusline(current_server, current_server_channel_nr, "PING to user %s timed out", server_list[current_server].user_ping);
					myfree(server_list[current_server].user_ping);
					server_list[current_server].user_ping = NULL;
				}
				else
				{
					update_statusline(current_server, current_server_channel_nr, "A PING to user %s is still in progress", server_list[current_server].user_ping);
					do_it = 0;
				}
			}

			if (!pars)
				popup_notify(FALSE, "/PING: you forgot to tell me what user to ping");
			else if (do_it)
			{
				server_list[current_server].user_ping = strdup(pars);
				server_list[current_server].t_user_ping = get_ts();
				server_list[current_server].user_ping_id = rand();

				if (irc_ping(server_list[current_server].fd, pars, server_list[current_server].user_ping_id) == -1)
					rc = -1;
			}
		}
		else if (strcasecmp(command, "IGNORE") == 0)
		{
			if (!pars)
				popup_notify(FALSE, "Ignore: nick name missing");
			else
			{
				int ni = find_nick(current_server, current_server_channel_nr, pars);

				if (ni != -1)
				{
					channel *pch = &server_list[current_server].pchannels[current_server_channel_nr];

					add_ignore(pch -> channel_name, pars, IGNORE_NOT_SET);

					if (pch -> persons[ni].ignored)
						popup_notify(FALSE, "IGNORE: already ignoring user %s!", pars);
					else
					{
						pch -> persons[ni].ignored = TRUE;
						update_statusline(current_server, current_server_channel_nr, "ignoring user %s", pars);
					}

					if (!save_ignore_list())
						popup_notify(FALSE, "Problem saving ignore-list file");
				}
				else
					popup_notify(FALSE, "IGNORE: user %s is not known", pars);
			}
		}
		else if (strcasecmp(command, "UNIGNORE") == 0)
		{
			if (!pars)
				popup_notify(FALSE, "Unignore: nick name missing");
			else
			{
				int ni = find_nick(current_server, current_server_channel_nr, pars);

				if (ni != -1)
				{
					channel *pch = &server_list[current_server].pchannels[current_server_channel_nr];

					del_ignore(pch -> channel_name, pars);

					if (pch -> persons[ni].ignored == FALSE)
						popup_notify(FALSE, "IGNORE: was not ignoring user %s?!", pars);
					else
					{
						pch -> persons[ni].ignored = FALSE;
						update_statusline(current_server, current_server_channel_nr, "no longer ignoring user %s", pars);
					}

					if (!save_ignore_list())
						popup_notify(FALSE, "Problem saving ignore-list file");
				}
				else
					popup_notify(FALSE, "IGNORE: user %s is not known", pars);
			}
		}
		else if (strcasecmp(command, "DCCSEND") == 0)
		{
			char *end_of_nick = pars ? strchr(pars, ' ') : NULL;
			if (!end_of_nick)
				popup_notify(FALSE, "DCCSEND: either the nickname or the filename are missing");
			else
			{
				*end_of_nick = 0x00;

				/* verify nick */
				if (find_nick(current_server, current_server_channel_nr, pars) == -1)
					popup_notify(FALSE, "DCCSEND: nick '%s' is not known for this channel", pars);
				else
				{
					if (init_send_dcc(end_of_nick + 1, current_server, pars) != 0)
						popup_notify(FALSE, "DCCSEND: failed to start transmitting file");
					else
						update_statusline(current_server, current_server_channel_nr, "DCCSEND: waiting for user '%s' to acknowledge file '%s'", pars, end_of_nick + 1);
				}
			}
		}
		else if (strcasecmp(command, "BAN") == 0)
		{
			if (!pars)
			{
				rc = do_send(server_list[current_server].fd, "MODE %s +b", server_list[current_server].pchannels[current_server_channel_nr].channel_name);
			}
			else
			{
				char *user = NULL;
				int loop;
				server *ps = &server_list[current_server];
				channel *pc = &ps -> pchannels[current_server_channel_nr];

				for(loop=0; loop<pc -> n_names; loop++)
				{
					if (strcasecmp((pc -> persons)[loop].name, pars) == 0)
					{
						user = (pc -> persons)[loop].name_complete;
						break;
					}
				}

				if (user)
				{
					if (irc_ban(ps -> fd, pc -> channel_name, user) == -1)
					{
						rc = -1;
					}
					else
					{
						update_statusline(current_server, current_server_channel_nr, "user %s (%s) banned", pars, user);
					}
				}
				else
				{
					popup_notify(FALSE, "BAN: user %s is not known", pars);
				}
			}
		}
		else if (strcasecmp(command, "RAW") == 0 || strcasecmp(command, "QUOTE") == 0)
		{
			if (do_send(server_list[current_server].fd, "%s", pars) == -1)
				rc = -1;
		}
		else if (strcasecmp(command, "LEAVE") == 0 || strcasecmp(command, "PART") == 0)
		{
			cmd_LEAVE(current_server, current_server_channel_nr, pars);
		}
		else if (strcasecmp(command, "TOPIC") == 0)
		{
			channel *pc = &server_list[current_server].pchannels[current_server_channel_nr];
			if (parsc == NULL || strlen(parsc) == 0)
			{
				char *topic = pc -> topic;

				update_statusline(current_server, current_server_channel_nr, "Topic for channel %s is: %s", pc -> channel_name, topic?topic: "not yet known.");
			}
			else if (irc_topic(server_list[current_server].fd, pc -> channel_name, parsc) == -1)
			{
				rc = -1;
			}
		}
		else if (strcasecmp(command, "KEEPTOPIC") == 0)
		{
			if (parsc == NULL || strlen(parsc) == 0)
			{
				myfree(server_list[current_server].pchannels[current_server_channel_nr].keeptopic);
				server_list[current_server].pchannels[current_server_channel_nr].keeptopic = NULL;

				update_statusline(current_server, current_server_channel_nr, "No longer maintaining topic on channel %s", server_list[current_server].pchannels[current_server_channel_nr].channel_name);
			}
			else
			{
				myfree(server_list[current_server].pchannels[current_server_channel_nr].keeptopic);
				server_list[current_server].pchannels[current_server_channel_nr].keeptopic = strdup(parsc);

				if (irc_topic(server_list[current_server].fd, server_list[current_server].pchannels[current_server_channel_nr].channel_name, parsc) == -1)
				{
					rc = -1;
				}
			}
		}
		else if (strcasecmp(command, "EXIT") == 0)	/* leave program */
		{
			exit_fi();
		}
		else if (strcasecmp(command, "QUIT") == 0)	/* leave server */
		{
			int server_index = -1;

			/* see if there's a channel given */
			if (pars && strlen(pars) > 0)
			{
				int loop;

				/* find server into server_index */
				for(loop=0; loop<n_servers; loop++)
				{
					if (strcasecmp(server_list[loop].server_host, pars) == 0)
					{
						server_index = loop;
						break;
					}
				}
			}

			if (server_index != -1)
			{
				if (n_servers == 1)
				{
					if (yesno_box(FALSE, "Terminate f-irc?", "Deleting the last/only server stops the program!", FALSE) == YES)
					{
						endwin();
						exit(0);
					}
				}
				else
				{
					close_server(server_index, TRUE);
					free_server(server_index);

					if (server_index <= (n_servers - 1))
						memmove(&server_list[server_index], &server_list[server_index + 1], sizeof(server) * (n_servers - (server_index + 1)));

					change_channel(0, 0, TRUE, FALSE);

					n_servers--;
				}
			}
			else
			{
				if (pars == NULL || strlen(pars) == 0)
					popup_notify(FALSE, "/QUIT is to leave a server (/EXIT to terminate the program)\nso it requires a parameter: the server(-name) to leave.");
				else
					popup_notify(FALSE, "Cannot leave server '%s': is not known", pars);
			}
		}
		else if (strcasecmp(command, "SPAM") == 0)	/* pm spam */
		{
			server *ps = &server_list[current_server];
			channel *pc = &ps -> pchannels[current_server_channel_nr];
			int loop;
			for(loop=0; loop<pc -> n_names; loop++)
			{
				if (irc_privmsg(ps -> fd, pc -> persons[loop].name, pars) == -1)
				{
					rc = -1;
				}
			}
		}
		else if (strcasecmp(command, "VERSIONSPAM") == 0)	/* ask all CTCP version */
		{
			int loop;
			server *ps = &server_list[current_server];
			channel *pc = &ps -> pchannels[current_server_channel_nr];

			for(loop=0; loop<pc -> n_names; loop++)
			{
				if (irc_privmsg(ps -> fd, pc -> persons[loop].name, "\001VERSION\001") == -1)
					rc = -1;
			}
		}
		else if (strcasecmp(command, "MSG") == 0 || strcasecmp(command, "QUERY") == 0)	/* private message */
		{
			if (!pars)
				popup_notify(FALSE, "%s: missing nick/channel and message", command);
			else
			{
				char *end_of_nick = strchr(pars, ' ');
				if (!end_of_nick)
					popup_notify(FALSE, "%s: missing nick/channel or message", command);
				else
				{
					*end_of_nick = 0x00;

					if (auto_private_channel)
					{
						int channel_nr = create_channel(current_server, pars);

						if (channel_nr == -1)
							rc = -1;
						else if (log_channel(current_server, channel_nr, server_list[current_server].nickname, end_of_nick + 1, FALSE) == -1)
							rc = -1;
					}

					if (irc_privmsg(server_list[current_server].fd, pars, end_of_nick + 1) == -1)
					{
						rc = -1;
					}
				}
			}
		}
		else if (strcasecmp(command, "CTCP") == 0)	/* CTCP message */
		{
			char *end_of_nick = pars ? strchr(pars, ' ') : NULL;
			if (!end_of_nick)
				popup_notify(FALSE, "CTCP: either the nickname or the message is missing");
			else
			{
				*end_of_nick = 0x00;

				/* verify nick */
				if (find_nick(current_server, current_server_channel_nr, pars) == -1)
					popup_notify(FALSE, "CTCP: nick '%s' is not known for this channel", pars);
				else
				{
					char *msg = NULL;
					asprintf(&msg, "\001%s\001", end_of_nick + 1);

					if (irc_privmsg(server_list[current_server].fd, pars, msg) == -1)
						rc = -1;

					myfree(msg);
				}
			}
		}
		else if (strcasecmp(command, "SAVECONFIG") == 0)
		{
			char *err_msg = NULL;

			(void)save_config(TRUE, &err_msg);
			popup_notify(FALSE, "%s", err_msg);

			free(err_msg);
		}
		else if (strcasecmp(command, "SEARCHALL") == 0)
			global_search(pars, pars);
		else
		{
			if (do_send(server_list[current_server].fd, "%s", user_line + 1) == -1)
				rc = -1;
		}

		myfree(command);
		myfree(parsc);
	}
	else			/* message to a certain channel */
	{
		char *line = NULL;

		if (me)
			asprintf(&line, "\001ACTION %s\001", &user_line[4]);
		else
			line = strdup(user_line);

		if (irc_privmsg(server_list[current_server].fd, server_list[current_server].pchannels[current_server_channel_nr].channel_name, line) == -1)
			rc = -1;

		if (log_channel(current_server, current_server_channel_nr, server_list[current_server].nickname, line, FALSE) == -1)
			rc = -1;

		myfree(line);
	}

	return rc;
}

void search_in_buffer_popup(const buffer *pb, const char *highlight)
{
	NEWWIN *bwin = NULL, *win = NULL;
	const char *what = edit_box(60, "^Search...^", NULL);
	char *title = NULL;
	buffer *search_results = NULL;

	if (!what)
		return; /* user selected cancel */

	if (strlen(what) == 0)
	{
		myfree(what);
		return; /* user pressed enter without nothing entered */
	}

	search_results = search_in_buffer_new(pb, what, FALSE);

	asprintf(&title, "Press left cursor key to exit, / to search | Searched %s", what);

	create_win_border(max_x - 4, max_y - 4, title, &bwin, &win);

	scrollback(win, search_results, what, FALSE, TRUE);

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	myfree(title);

	free_buffer(search_results);

	myfree(what);
}

void search_everywhere(void)
{
	NEWWIN *bwin = NULL, *win = NULL;
	const char *what = edit_box(60, "^Search all ...^", NULL);
	char *title = NULL;
	int sr = 0;
	buffer *search_results = NULL;

	if (!what || strlen(what) == 0)
	{
		myfree(what);
		return; /* user selected cancel or enter with nothing entered */
	}

	search_results = create_buffer(0);

	for(sr=0; sr<n_servers; sr++)
	{
		server *ps = &server_list[sr];
		int ch = 0;

		for(ch=0; ch<ps -> n_channels; ch++)
		{
			channel *pc = &ps -> pchannels[ch];

			search_in_buffer(pc -> pbuffer, search_results, what, TRUE);
		}
	}

	asprintf(&title, "Press left cursor key to exit, / to search | Searched for %s", what);

	create_win_border(max_x - 4, max_y - 4, title, &bwin, &win);

	scrollback(win, search_results, what, FALSE, TRUE);

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	myfree(title);

	free_buffer(search_results);

	myfree(what);
}

int user_menu(int server_index, int channel_index, int name_index)
{
	int rc = 0;
	NEWWIN *bwin = NULL, *win = NULL;
	channel *pc = &server_list[server_index].pchannels[channel_index];
	char *channel_name = pc -> channel_name;
	const char *nick = NULL, *name_complete = NULL;
	char *title = NULL;

	if (name_index < 0 || name_index > pc -> n_names)
	{
		delete_window(win);
		delete_window(bwin);
		mydoupdate();

		return 0;
	}

	nick = pc -> persons[name_index].name;
	name_complete = pc -> persons[name_index].name_complete;

	asprintf(&title, "%s [%s] (%s)", nick, channel_name, server_list[server_index].server_host);

	create_win_border(48, 17, title, &bwin, &win);

	free(title);

	mywattron(win -> win, A_BOLD);
	mvwprintw(win -> win, 1, 2, "%s", name_complete);
	mywattroff(win -> win, A_BOLD);

	escape_print_xy(win, 2, 2, "^k^ kick user, ^K^ kick with message");
	escape_print_xy(win, 3, 2, "^b^ ban nick,  ^B^ ban+kick nick");
	escape_print_xy(win, 4, 2, "^g^ ban user,  ^G^ ban+kick user");
	escape_print_xy(win, 5, 2, "^o^ op,        ^O^ de-op");
	escape_print_xy(win, 6, 2, "^w^ whois");
	escape_print_xy(win, 7, 2, "^v^ version");
	escape_print_xy(win, 8, 2, "^a^ allow speak");
	escape_print_xy(win, 9, 2, "^A^ disallow speak");
	escape_print_xy(win, 10, 2, "^p^ ping!");
	escape_print_xy(win, 11, 2, "^i^ ignore,   ^I^ un-ignore");
	escape_print_xy(win, 12, 2, "^D^ send a file using DCC");
	escape_print_xy(win, 13, 2, "^c^ CTCP");
	escape_print_xy(win, 14, 2, "^q^/^LEFT cursor key^ exit this menu");

	if (pc -> persons[name_index].ignored)
		escape_print_xy(win, 15, 2, "currently ignoring user");

	mydoupdate();

	for(;;)
	{
		int c = wait_for_keypress(FALSE);

		if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == 'q')
			break;
		else if (c == 3)
			exit_fi();
		else if (c == 'i')
		{
			channel *pc = &server_list[server_index].pchannels[channel_index];

			pc -> persons[name_index].ignored = TRUE;

			add_ignore(pc -> channel_name, pc -> persons[name_index].name, pc -> persons[name_index].name_complete);

			if (!save_ignore_list())
				popup_notify(FALSE, "Problem saving ignore-list file");

			if (rc == 0 && log_channel(server_index, channel_index, nick, "ignore", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'I')
		{
			channel *pc = &server_list[server_index].pchannels[channel_index];

			pc -> persons[name_index].ignored = FALSE;

			del_ignore(pc -> channel_name, pc -> persons[name_index].name);

			if (!save_ignore_list())
				popup_notify(FALSE, "Problem saving ignore-list file");

			if (rc == 0 && log_channel(server_index, channel_index, nick, "unignore", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'k' || c == 'K')
		{
			const char *msg = NULL;
			BOOL dokick = TRUE;

			if (c == 'K')
			{
				msg = edit_box(60, "^Kick message^", NULL);

				if (!msg)
					dokick = FALSE;
			}

			if (dokick)
			{
				rc = irc_kick(server_list[server_index].fd, channel_name, nick, msg);

				if (rc == 0 && log_channel(server_index, channel_index, nick, "kick", TRUE) == -1)
					rc = -1;
			}

			myfree(msg);

			break;
		}
		else if (c == 'b' || c == 'g')
		{
			const char *what = c == 'b' ? nick : name_complete;

			rc = irc_ban(server_list[server_index].fd, channel_name, what);

			if (rc == 0 && log_channel(server_index, channel_index, what, "BAN", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'B' || c == 'G')
		{
			const char *msg = edit_box(60, "^Kick+ban message^", NULL);

			if (msg)
			{
				const char *what = c == 'B' ? nick : name_complete;

				rc = irc_ban(server_list[server_index].fd, channel_name, what);

				if (rc == -1)
					update_statusline(server_index, channel_index, "BAN failed: also not kicked\n");
				else
				{
					rc = irc_kick(server_list[server_index].fd, channel_name, what, msg);

					if (rc == 0 && log_channel(server_index, channel_index, what, "BAN + KICK", TRUE) == -1)
						rc = -1;
				}

				myfree(msg);
			}

			break;
		}
		else if (c == 'o')
		{
			rc = irc_op(server_list[server_index].fd, channel_name, nick, TRUE);

			if (rc == 0 && log_channel(server_index, channel_index, nick, "set OP", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'O')
		{
			rc = irc_op(server_list[server_index].fd, channel_name, nick, FALSE);

			if (rc == 0 && log_channel(server_index, channel_index, nick, "remove OP", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'w')
		{
			rc = irc_whois(server_list[server_index].fd, nick);

			if (rc == 0 && log_channel(server_index, 0, nick, "CTCP WHOIS", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'v')
		{
			rc = irc_privmsg(server_list[server_index].fd, nick, "\001VERSION\001");

			if (rc == 0 && log_channel(server_index, 0, nick, "CTCP VERSION", TRUE) == -1)
				rc = -1;
			break;
		}
		else if (c == 'a')
		{
			popup_notify(FALSE, "Speak allowed");

			rc = irc_allowspeak(server_list[server_index].fd, channel_name, nick, TRUE);

			if (rc == 0 && log_channel(server_index, channel_index, nick, "allow speak", TRUE) == -1)
				rc = -1;
		}
		else if (c == 'A')
		{
			popup_notify(FALSE, "Speak disabled");

			rc = irc_allowspeak(server_list[server_index].fd, channel_name, nick, FALSE);

			if (rc == 0 && log_channel(server_index, channel_index, nick, "disallow speak", TRUE) == -1)
				rc = -1;
		}
		else if (c == 'p')
		{
			popup_notify(FALSE, "Ping sent");

			server_list[server_index].user_ping = strdup(nick);
			server_list[server_index].t_user_ping = get_ts();
			server_list[server_index].user_ping_id = rand();

			rc = irc_ping(server_list[server_index].fd, nick, server_list[server_index].user_ping_id);

			if (rc == 0 && log_channel(server_index, 0, nick, "send ping", TRUE) == -1)
				rc = -1;
		}
		else if (c == 'D')
		{
			const char *filename_in = edit_box(60, "^DCC send file^", NULL);

			rc = 0;
			if (filename_in)
			{
				char *log_msg = NULL;
				const char *filename = realpath(filename_in, NULL);

				rc = init_send_dcc(filename, server_index, nick);

				asprintf(&log_msg, "DCC send %s", filename);

				if (rc == 0 && log_channel(server_index, 0, nick, log_msg, TRUE) == -1)
					rc = -1;

				myfree(filename);

				myfree(log_msg);

				myfree(filename_in);
			}
		}
		else if (c == 'c')
		{
			const char *ctcp_command = edit_box(60, "^CTCP^", NULL);

			rc = 0;
			if (ctcp_command)
			{
				char *msg = NULL, *log_msg = NULL;

				asprintf(&msg, "\001%s\001", ctcp_command);

				rc = irc_privmsg(server_list[server_index].fd, nick, msg);

				asprintf(&log_msg, "CTCP: %s", ctcp_command);

				if (rc == 0 && log_channel(server_index, 0, nick, log_msg, TRUE) == -1)
					rc = -1;

				myfree(log_msg);

				myfree(msg);

				myfree(ctcp_command);
			}
		}
		else
			wrong_key();
	}

	delete_window(win);
	delete_window(bwin);
	mydoupdate();

	return rc;
}

void refresh_window_with_buffer(NEWWIN *where, const int window_height, buffer *pbuffer, const char *hl, BOOL force_partial_highlight)
{
	int loop=0, n_elements = get_buffer_n_elements(pbuffer);

	werase(where -> win);

	for(loop=max(n_elements - window_height, 0); loop<n_elements; loop++)
	{
		nick_color_settings ncs;

		find_nick_colorpair(get_from_buffer(pbuffer, loop) -> msg_from, &ncs);

		output_to_window(where, get_from_buffer(pbuffer, loop) -> msg, hl, get_from_buffer(pbuffer, loop) -> meta, nick_color ? &ncs : NULL, force_partial_highlight);
	}

	pbuffer -> last_shown = n_elements;

	mydoupdate();
}

char *generate_line(buffer_element_t *line, BOOL with_server, BOOL with_channel, int max_width)
{
	if (with_channel)
	{
		char *out = NULL;
		const char *c_name = server_list[line -> sr].pchannels[line -> ch].channel_name;

		if (with_server)
		{
			const char *s_name = server_list[line -> sr].description;
			if (s_name == NULL)
				s_name = server_list[line -> sr].server_host;

			asprintf(&out,  "%s|%s] %s", s_name, c_name, line -> msg);
		}
		else
		{
			asprintf(&out,  "%s] %s", c_name, line -> msg);
		}

		if (max_width > 0 && strlen(out) > max_width)
			out[max_width] = 0x00;

		return out;
	}

	return strdup(line -> msg);
}

void write_buffer_to_file(const buffer *pb, BOOL with_channel)
{
	const char *file = edit_box(60, "^Select file to write to...^", NULL);

	if (file)
	{
		FILE *fh = fopen(file, "wb");
		if (!fh)
			popup_notify(FALSE, "Failed to create file: %s", strerror(errno));
		else
		{
			BOOL err = FALSE;
			int loop = 0;

			for(loop=0; loop<get_buffer_n_elements(pb); loop++)
			{
				buffer_element_t *line = get_from_buffer(pb, loop);

				char *msg = generate_line(line, n_servers > 1, with_channel, -1);

				err |= fprintf(fh, "%s\n", msg) < 1;

				free(msg);
			}

			fclose(fh);

			if (err)
				popup_notify(FALSE, "An error occured writing to file %s", file);
		}

		myfree(file);
	}
}

void scrollback_displayline(NEWWIN *win, buffer_element_t *line, const int terminal_offset, const char *highlight, BOOL with_channel, int max_width, BOOL force_partial_highlight)
{
	nick_color_settings ncs;
	char * out = generate_line(line, n_servers > 1, with_channel, max_width);

	wmove(win -> win, terminal_offset, 0);

	find_nick_colorpair(line -> msg_from, &ncs);

	output_to_window(win, out, highlight, line -> meta, nick_color ? &ncs : NULL, force_partial_highlight);

	myfree(out);
}

void global_search(const char *highlight, const char *search_what)
{
	NEWWIN *bwin = NULL, *win = NULL;
	const char *what = search_what ? strdup(search_what) : edit_box(60, "^Global search...^", NULL);
	char *title = NULL;
	buffer *search_results = create_buffer(2500);
	int s = 0, c = 0;

	if (!what)
	{
		free_buffer(search_results);

		return; /* user selected cancel */
	}

	if (strlen(what) == 0)
	{
		myfree(what);

		free_buffer(search_results);

		return; /* user pressed enter without nothing entered */
	}

	asprintf(&title, "Press left cursor key to exit, / to search | Global search for %s", what);

	create_win_border(max_x - 4, max_y - 4, title, &bwin, &win);

	for(s=0; s<n_servers; s++)
	{
		for(c=0; c<server_list[s].n_channels; c++)
			search_in_buffer(server_list[s].pchannels[c].pbuffer, search_results, what, TRUE);
	}

	sort_buffer(search_results, TRUE);

	scrollback(win, search_results, highlight, TRUE, TRUE);

	delete_window(win);
	delete_window(bwin);
	mydoupdate();

	myfree(title);

	free_buffer(search_results);

	myfree(what);
}

void scrollback_new(const buffer *pbuffer, const char *highlight, const char *title_in, BOOL with_channel, BOOL force_partial_highlight)
{
	char *title = NULL;
	NEWWIN *bwin = NULL, *win = NULL;

	asprintf(&title, "Press left cursor key to exit, / to search | %s", title_in);

	create_win_border(max_x - 4, max_y - 4, title, &bwin, &win);

	scrollback(win, pbuffer, highlight, with_channel, force_partial_highlight);

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	myfree(title);
}

void scrollback(NEWWIN *win, const buffer *pbuffer, const char *highlight, BOOL with_channel, BOOL force_partial_highlight)
{
	int offset = max(0, get_buffer_n_elements(pbuffer) - win -> nlines);
	char redraw = 2;

	for(;;)
	{
		int c = 0;

		if (redraw == 2)
		{
			int index = 0;
			int lines_used = 0;

			werase(win -> win);

			for(;offset + index < get_buffer_n_elements(pbuffer);)
			{
				buffer_element_t *pbe = get_from_buffer(pbuffer, offset + index);

				int prev_lines_used = lines_used;

				lines_used += (strlen(pbe -> msg) + win -> ncols - 1) / win -> ncols;
				if (lines_used > win -> nlines)
					break;

				scrollback_displayline(win, pbe, prev_lines_used, highlight, with_channel, -1, force_partial_highlight);

				index++;
			}

			redraw = 1;
		}

		if (redraw)
		{
			mydoupdate();

			redraw = 0;
		}

		c = wait_for_keypress(FALSE);

		if (toupper(c) == 'Q' || toupper(c) == 'X')
			break;
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;
		else if (c == 3)
			exit_fi();
		else if (c == 'w') /* write to file */
		{
			write_buffer_to_file(pbuffer, with_channel);
		}
		else if (c == '/') /* search */
		{
			search_in_buffer_popup(pbuffer, highlight);
		}
		else if (c == KEY_RIGHT)
		{
			wrong_key();
		}
		else if (c == KEY_UP && offset > 0)
		{
			int n_to_move = 0;
			buffer_element_t *pbe = NULL;

			offset--;

			wmove(win -> win, 0, 0);

			pbe = get_from_buffer(pbuffer, offset);

			n_to_move = (strlen(pbe -> msg) + win -> ncols - 1) / win -> ncols;

			winsdelln(win -> win, n_to_move);

			scrollback_displayline(win, pbe, 0, highlight, with_channel, -1, force_partial_highlight);

			redraw = 1;
		}
		else if ((c == KEY_DOWN || c == KEY_NPAGE) && offset < get_buffer_n_elements(pbuffer) - 1)
		{
			int dummy = 0;
			while(offset < get_buffer_n_elements(pbuffer) - 1)
			{
				offset++;

				dummy += (strlen(get_from_buffer(pbuffer, offset) -> msg) + win -> ncols - 1) / win -> ncols;

				if ((dummy > win -> nlines && c == KEY_NPAGE) ||
				    (dummy > win -> nlines / 4 && c == KEY_DOWN))
				{
					offset--;
					break;
				}
			}

			redraw = 2;
		}
		else if (c == KEY_PPAGE && offset > 0)
		{
			int dummy = 0;
			while(offset > 0)
			{
				offset--;

				dummy += (strlen(get_from_buffer(pbuffer, offset) -> msg) + win -> ncols - 1) / win -> ncols;
				if (dummy > win -> nlines)
				{
					offset++;
					break;
				}
			}

			redraw = 2;
		}
		else if (c == KEY_HOME && offset > 0)
		{
			offset = 0;
		}
		else if (c == KEY_END && offset < get_buffer_n_elements(pbuffer) - 1)
		{
			offset = get_buffer_n_elements(pbuffer) - 1;
		}
		else
		{
			wrong_key();
		}
	}
}

int select_server(void)
{
	NEWWIN *bwin = NULL, *win = NULL;
	char *title = NULL;
	int lines = 12, cols = 59;
	int offset = 0, woffset = 0, selection = -1;

	asprintf(&title, "Select server");

	create_win_border(63, 16, title, &bwin, &win);

	free(title);

	for(;;)
	{
		int c = -1, index = 0;

		werase(win -> win);

		for(index=0; index<lines && index + offset < n_servers; index++)
		{
			char *buffer = (char *)malloc(cols + 1);
			char *temp = NULL;
			int si = index + offset;
			int clen = asprintf(&temp, "%s (%s:%d)", server_list[si].description, server_list[si].server_host, server_list[si].server_port);

			memset(buffer, ' ', cols);
			buffer[cols] = 0x00;

			memcpy(buffer, temp, clen);

			free(temp);

			if (index == woffset)
				mywattron(win -> win, A_REVERSE);

			mvwprintw(win -> win, 1 + index, 2, buffer);

			if (index == woffset)
				mywattroff(win -> win, A_REVERSE);

			free(buffer);
		}

		mydoupdate();

		c = wait_for_keypress(TRUE);

		if (c == 0)
			continue;

		if (c == 'q' || c == 'Q' || c == -1)
			break;
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == KEY_UP)
		{
			if (woffset > 0)
				woffset--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();
		}
		else if (c == KEY_DOWN)
		{
			if (woffset < lines - 1 && woffset + offset < n_servers - 1)
				woffset++;
			else if (offset + woffset < n_servers - 1)
				offset++;
			else
				wrong_key();
		}
		else if (c == 3)
		{
			if (yesno_box(FALSE, "Terminate f-irc", "Are you sure you want to terminate the program?", FALSE) == YES)
				exit_fi();
		}
		else if (c == ' ' || c == KEY_RIGHT || c == 13)
		{
			selection = offset + woffset;
			break;
		}
		else
		{
			wrong_key();
		}
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	return selection;
}

void edit_send_after_connect(int sr)
{
	server *ps = &server_list[sr];
	char *title = NULL;

	asprintf(&title, "EDIT %s \"send after connect\", left cursor to exit", ps -> server_host);

	edit_string_array(&ps -> send_after_login, title);

	free(title);
}

void edit_auto_join(int sr)
{
	server *ps = &server_list[sr];
	char *title = NULL;

	asprintf(&title, "EDIT %s \"auto-join\", left cursor to exit", ps -> server_host);

	edit_string_array(&ps -> auto_join, title);

	free(title);
}

void edit_favorites(void)
{
	int i = 0;
	string_array_t f;

	init_string_array(&f);

	for(i=0; i<n_favorite_channels; i++)
	{
		char *line = NULL;

		if (favorite_channels[i].server)
			asprintf(&line, "%s %s", favorite_channels[i].server, favorite_channels[i].channel);
		else
			asprintf(&line, "%s", favorite_channels[i].channel);

		add_to_string_array(&f, line);

		free(line);
	}

	edit_string_array(&f, "EDIT favorites");

	free_favorites();

	for(i=0; i<string_array_get_n(&f); i++)
	{
		int n = -1;
		string_array_t parts;
		init_string_array(&parts);

		split_string(string_array_get(&f, i), " ", TRUE, &parts);

		n = string_array_get_n(&parts);

		if (n == 2)
			add_favorite(string_array_get(&parts, 0), string_array_get(&parts, 1));
		else if (n == 1)
			add_favorite(NULL, string_array_get(&parts, 0));

		free_string_array(&parts);
	}

	favorite_channels_index = 0;

	free_string_array(&f);
}

void edit_string_array(string_array_t *p, const char *title)
{
	NEWWIN *bwin = NULL, *win = NULL;
	int lines = 9, cols = 59;
	int offset = 0, woffset = 0;

	create_win_border(63, 16, title, &bwin, &win);

	for(;;)
	{
		int c = -1, index = 0;

		werase(win -> win);

		escape_print_xy(win, 1, 2, "^a^ add  ^e^/^right cursor key^ edit  ^i^ insert");
		escape_print_xy(win, 2, 2, "^d^ del (the one under the cursor)");

		for(index=0; index<lines && offset + index < string_array_get_n(p); index++)
		{
			char *buffer = (char *)malloc(cols + 1);
			int use_index = index + offset;
			int clen = strlen(string_array_get(p, use_index));

			memset(buffer, ' ', cols);
			buffer[cols] = 0x00;

			memcpy(buffer, string_array_get(p, use_index), clen);

			if (index == woffset)
				mywattron(win -> win, A_REVERSE);

			mvwprintw(win -> win, 4 + index, 2, buffer);

			if (index == woffset)
				mywattroff(win -> win, A_REVERSE);

			free(buffer);
		}

		mydoupdate();

		c = wait_for_keypress(TRUE);

		if (c == 0)
			continue;

		if (c == 'q' || c == 'Q' || c == -1)
			break;
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == KEY_UP)
		{
			if (woffset > 0)
				woffset--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();
		}
		else if (c == KEY_DOWN)
		{
			if (woffset < lines - 1 && woffset + offset < string_array_get_n(p) - 1)
				woffset++;
			else if (woffset + offset < string_array_get_n(p) - 1)
				offset++;
			else
				wrong_key();
		}
		else if (c == 3)
		{
			if (yesno_box(FALSE, "Terminate f-irc", "Are you sure you want to terminate the program?", FALSE) == YES)
				exit_fi();
		}
		else if (c == 'd')
		{
			if (yesno_box(FALSE, "Delete", "Are you sure you want to delete this entry?", FALSE) == YES)
				del_nr_from_string_array(p, offset + woffset);
		}
		else if (c == 'a')
		{
			char *n = (char *)edit_box(60, "^Add^", NULL);

			if (n != NULL)
			{
				add_to_string_array(p, n);

				free(n);
			}
		}
		else if (c == 'i')
		{
			char *n = (char *)edit_box(60, "^Insert^", NULL);

			if (n != NULL)
			{
				insert_into_string_array(p, offset + woffset, n);

				free(n);
			}
		}
		else if (c == 'e' || c == KEY_RIGHT || c == 13)
		{
			char *n = (char *)edit_box(60, "^Edit^", string_array_get(p, offset + woffset));

			if (n != NULL)
			{
				replace_in_string_array(p, offset + woffset, n);

				free(n);
			}
		}
		else
		{
			wrong_key();
		}
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();
}

void edit_server(int sr)
{
	NEWWIN *bwin = NULL, *win = NULL;
	server *ps = &server_list[sr];
	char *title = NULL;
	BOOL restart_required = FALSE;

	asprintf(&title, "EDIT %s:%d, left cursor key to exit", ps -> server_host, ps -> server_port);

	create_win_border(63, 16, title, &bwin, &win);

	free(title);

	for(;;)
	{
		int c = -1;

		werase(win -> win);

		escape_print_xy(win, 1, 2, "^d^ description");
		escape_print_xy(win, 2, 2, "^h^ host");
		escape_print_xy(win, 3, 2, "^p^ port");
		escape_print_xy(win, 4, 2, "^U^ complete name");
		escape_print_xy(win, 5, 2, "^u^ username");
		escape_print_xy(win, 6, 2, "^P^ password");
		escape_print_xy(win, 7, 2, "^n^ nickname");
		escape_print_xy(win, 8, 2, "^S^ edit \"send after connect\"");
		escape_print_xy(win, 9, 2, "^A^ edit \"auto-join\"");

		mydoupdate();

		c = wait_for_keypress(TRUE);

		if (c == 0)
			continue;

		if (c == 'q' || c == 'Q' || c == -1)
			break;
		if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == 3)
		{
			if (yesno_box(FALSE, "Terminate f-irc", "Are you sure you want to terminate the program?", FALSE) == YES)
				exit_fi();
		}
		else if (c == 'd')
		{
			const char *n = edit_box(60, "^Edit server description^", ps -> description);

			if (n != NULL)
			{
				myfree(ps -> description);
				ps -> description = n;
			}
		}
		else if (c == 'h')
		{
			const char *n = edit_box(60, "^Edit server host^", ps -> server_host);

			if (n != NULL)
			{
				myfree(ps -> server_host);
				ps -> server_host = n;
			}

			restart_required = TRUE;
		}
		else if (c == 'p')
		{
			char *port_str = NULL;
			const char *n = NULL;

			asprintf(&port_str, "%d", ps -> server_port);
			n = edit_box(60, "^Edit server port^", port_str);
			free(port_str);

			if (n != NULL)
				ps -> server_port = atoi(n);

			restart_required = TRUE;
		}
		else if (c == 'U')
		{
			char *n = (char *)edit_box(60, "^Edit \"complete name\"^", ps -> user_complete_name);

			if (n != NULL)
			{
				free(ps -> user_complete_name);
				ps -> user_complete_name = n;
			}

			restart_required = TRUE;
		}
		else if (c == 'u')
		{
			char *n = (char *)edit_box(60, "^Edit user name^", ps -> username);

			if (n != NULL)
			{
				free(ps -> username);
				ps -> username = n;
			}

			restart_required = TRUE;
		}
		else if (c == 'P')
		{
			char *n = (char *)edit_box(60, "^Edit password^", ps -> password);

			if (n != NULL)
			{
				free(ps -> password);
				ps -> password = n;
			}

			restart_required = TRUE;
		}
		else if (c == 'n')
		{
			char *n = (char *)edit_box(60, "^Edit nick^", ps -> nickname);

			if (n != NULL)
			{
				free(ps -> nickname);
				ps -> nickname = n;
			}

			restart_required = TRUE;
		}
		else if (c == 'S')
			edit_send_after_connect(sr);
		else if (c == 'A')
			edit_auto_join(sr);
		else
		{
			wrong_key();
		}
	}

	if (restart_required)
		popup_notify(FALSE, "The changes you made require you to\n\"restart the irc-connection\".\nTo do so, press 'r' in the server menu.");

	delete_window(win);
	delete_window(bwin);

	mydoupdate();
}

void server_menu(int sr)
{
	NEWWIN *bwin = NULL, *win = NULL;
	server *ps = &server_list[sr];
	char *title = NULL;

	asprintf(&title, "%s [%s]", ps -> description, ps -> server_host);

	create_win_border(63, 16, title, &bwin, &win);

	free(title);

	for(;;)
	{
		char *ss = NULL;
		int c = -1;
		char *ch_list_compl = NULL;
		char *addr = NULL, *addr_real = NULL, *bps_str = NULL;
		time_t now = time(NULL);
		double bps = 0.0;
		int t_diff = now - ps -> ts_bytes;

		werase(win -> win);

		asprintf(&addr, "address: ^[%s]:%d^", ps -> server_host, ps -> server_port);
		escape_print_xy(win, 3, 2, addr);
		free(addr);

		asprintf(&addr_real, "connected to: ^%s^", ps -> server_real);
		escape_print_xy(win, 4, 2, addr_real);
		free(addr_real);

		bps = ps -> prev_bps;

		if (t_diff)
			bps = (bps + ps -> bytes / (double)t_diff) / 2;

		asprintf(&bps_str, "BPS: %.2f", bps);

		escape_print_xy(win, 5, 2, bps_str);
		free(bps_str);

		escape_print_xy(win, 7, 2, "^r^ reconnect");
		escape_print_xy(win, 8, 2, "^p^ disconnect & remove");
		escape_print_xy(win, 9, 2, "^g^ (re-)retrieve channel list (/LIST)");
		escape_print_xy(win, 10, 2, "^j^ join a channel");
		escape_print_xy(win, 11, 2, "^e^ edit server settings");
		escape_print_xy(win, 14, 2, "^q^/^LEFT cursor key^ exit this menu");

		if (ps -> all_channels_complete == TRUE)
			asprintf(&ch_list_compl, "List of channels available (%d known)", ps -> n_all_channels);
		else
			asprintf(&ch_list_compl, "List of channels NOT YET available (received: %d)", ps -> n_all_channels);

		escape_print_xy(win, 1, 2, ch_list_compl);

		free(ch_list_compl);

		switch(ps -> state)
		{
			case STATE_NO_CONNECTION:
				ss = "Server state: no connection, will connect";
				break;
			case STATE_ERROR:
				ss = "Server state: connect error";
				break;
			case STATE_TCP_CONNECT:
				ss = "Server state: connecting (TCP)";
				break;
			case STATE_IRC_CONNECTING:
				ss = "Server state: connecting (IRC)";
				break;
			case STATE_CONNECTED1:
				ss = "Server state: connecting (IRC handshake phase 1)";
				break;
			case STATE_CONNECTED2:
				ss = "Server state: connecting (IRC handshake phase 2)";
				break;
			case STATE_LOGGING_IN:
				ss = "Server state: connecting (logging in)";
				break;
			case STATE_RUNNING:
				ss = "Server state: connected!";
				break;
			case STATE_DISCONNECTED:
				ss = "Server state: disconnected";
				break;
			default:
				ss = "Server state: UNKNOWN!";
				break;
		}

		escape_print_xy(win, 2, 2, ss);

		mydoupdate();

		c = wait_for_keypress(TRUE);

		if (c == 0)
			continue;

		if (c == 'q' || c == 'Q' || c == -1)
			break;
		if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == 3)
			exit_fi();
		else if (c ==  'p')
		{
			if (n_servers > 1)
			{
				if (yesno_box(FALSE, "Remove server", "Are you sure?", FALSE) == YES)
				{
					close_server(sr, TRUE);
					free_server(sr);

					if (sr <= n_servers - 1)
						memmove(&server_list[sr], &server_list[sr + 1], sizeof(server) * (n_servers - (sr + 1)));

					n_servers--;

					show_channel_names_list();
					change_channel(0, 0, TRUE, FALSE);

					popup_notify(FALSE, "Server removed");
				}
			}
			else
			{
				popup_notify(FALSE, "Cannot remove server: need at least 1.");
			}

			break;
		}
		else if (c == 'r')
		{
			server_list[sr].reconnect_delay = DEFAULT_RECONNECT_DELAY;

			restart_server(sr);

			popup_notify(FALSE, "Restarting connection");
		}
		else if (c == 'g')
		{
			if (irc_list(ps -> fd) == -1)
			{
				popup_notify(FALSE, "Disconnected from server");
				break;
			}
			else
			{
				popup_notify(FALSE, "Requested channel list");
			}
		}
		else if (c == 'e')
		{
			edit_server(sr);
		}
		else if (c == 'j')
		{
			if (ps -> all_channels_complete == FALSE)
				popup_notify(FALSE, "Please wait for the channel list to load");
			else
			{
				buffer *temp = create_buffer(server_list[sr].n_all_channels);
				int loop, csr = -1, cch = -1;
				char *dummy = NULL;
				BOOL err = FALSE;
				NEWWIN *bwin = NULL, *win = NULL;

				create_win_border(40, max_y - 4, "right cursor key: select, left: cancel", &bwin, &win);

				for(loop=0; loop<server_list[sr].n_all_channels; loop++)
					add_to_buffer(temp, server_list[sr].all_channels[loop], NULL, FALSE, sr, loop);

				if (scrollback_and_select(win, temp, &dummy, &csr, &cch, server_list[sr].nickname, FALSE, FALSE))
				{
					if (irc_join(server_list[sr].fd, get_from_buffer(temp, cch) -> msg))
						err = TRUE;
				}

				delete_window(win);
				delete_window(bwin);

				mydoupdate();

				myfree(dummy);

				free_buffer(temp);

				if (err)
					break;
			}
		}
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();
}

BOOL scrollback_and_select(NEWWIN *win, const buffer *pbuffer, char **sel_str, int *sr, int *ch, const char *needle, BOOL with_channel, BOOL force_partial_highlight)
{
	int ppos = -1;
	int offset = 0, cursor = 0;

	for(;;)
	{
		int c;

		/* redraw */
		if (offset + cursor != ppos)
		{
			int loop;

			werase(win -> win);

			for(loop=0; loop < min(get_buffer_n_elements(pbuffer), win -> nlines); loop++)
			{
				if (loop == cursor)
					mywattron(win -> win, A_REVERSE);

				scrollback_displayline(win, get_from_buffer(pbuffer, offset + loop), loop, needle, with_channel, win -> ncols, force_partial_highlight);

				if (loop == cursor)
					mywattroff(win -> win, A_REVERSE);
			}

			mydoupdate();

			ppos = offset + cursor;
		}

		c = wait_for_keypress(FALSE);

		if (tolower(c) == 'q')
			return FALSE;
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			return FALSE;
		else if (c == KEY_RIGHT || c == 13)
		{
			if (offset + cursor < get_buffer_n_elements(pbuffer))
			{
				buffer_element_t *pbe = get_from_buffer(pbuffer, offset + cursor);

				*sel_str = strdup(pbe -> msg);
				*sr = pbe -> sr;
				*ch = pbe -> ch;

				return TRUE;
			}

			wrong_key();
		}
		else if (c == 'w') /* write to file */
		{
			write_buffer_to_file(pbuffer, TRUE);
		}
		else if (c == '/') /* search */
		{
			search_in_buffer_popup(pbuffer, needle);
		}
		else if (c == KEY_HOME)
		{
			offset = cursor = 0;
		}
		else if (c == KEY_END)
		{
			int n_lines = get_buffer_n_elements(pbuffer);

			if (n_lines >= win -> nlines)
			{
				cursor = win -> nlines - 1;
				offset = n_lines - win -> nlines;
			}
			else
			{
				cursor = n_lines - 1;
				offset = 0;
			}

		}
		else if (c == KEY_UP)
		{
			if (cursor > 0)
				cursor--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();
		}
		else if (c == KEY_DOWN)
		{
			if (cursor + offset < get_buffer_n_elements(pbuffer) - 1)
			{
				if (cursor < win -> nlines - 1)
					cursor++;
				else
					offset++;
			}
			else
			{
				wrong_key();
			}
		}
		else if (c == KEY_PPAGE)
		{
			if (cursor)
				cursor = 0;
			else if (offset >= win -> nlines -1)
				offset -= win -> nlines - 1;
			else
				wrong_key();
		}
		else if (c == KEY_NPAGE)
		{
			if (cursor + offset < get_buffer_n_elements(pbuffer) - (win -> nlines - 1))
				offset += win -> nlines - 1;
			else
				wrong_key();
		}
		else if (c == 3)
		{
			exit_fi();
		}
		else
		{
			wrong_key();
		}
	}

	return FALSE;
}

void online_help(void)
{
	const int help_win_width = 77, help_win_height = 17;
	NEWWIN *bwin = NULL, *win = NULL;
	int screen_nr = 0;

	create_win_border(help_win_width, help_win_height, "left cursor key: previous screen / exit help, right key: next screen", &bwin, &win);

	for(;;)
	{
		int c = 0;

		werase(win -> win);

		if (screen_nr == 0)
		{
			escape_print(win, "\n", '|');
			escape_print(win, " Press |right cursor key| to:\n", '|');
			escape_print(win, "  - enter a screen\n", '|');
			escape_print(win, "  - unfold a server (=show channels)\n", '|');
			escape_print(win, "  - show a server commands menu\n", '|');
			escape_print(win, "  - when in a channel: show list of people (nicks)\n", '|');
			escape_print(win, "  - when in a list of people: show a user menu\n", '|');
			escape_print(win, "  - in the word cloud: do a search for the selected word\n", '|');
			escape_print(win, " \n", '|');
			escape_print(win, " Press |left cursor key| to:\n", '|');
			escape_print(win, "  - leave a screen\n", '|');
			escape_print(win, "  - fold a server (=hide channels)\n", '|');
			escape_print(win, "  - go back from the people list to the channel list\n", '|');
			escape_print(win, " \n", '|');
			escape_print(win, " Press right cursor key to go to the next help-screen\n", '|');
			escape_print(win, " which lists the shortcut keys.\n", '|');
		}
		else if (screen_nr == 1)
		{
			escape_print(win, "\n", '|');
			escape_print(win, " |^A| moves the cursor to the left |^E| move right |^D| deletes current character\n", '|');
			escape_print(win, " |^U| clear line, enter twice to undo clear\n", '|');
			escape_print(win, " |^W| jump to the next channel with new text - ^R jump backwards\n", '|');
			escape_print(win, " |^Z| jump to the next channel with your nick name - ^Z search backwards\n", '|');
			escape_print(win, " |^Q| jump to next favorite channel (see configuration file)\n", '|');
			escape_print(win, " |^O| remove all the *'s (forget what channels have unread text)\n", '|');
			escape_print(win, " |^T| go back to the previous selected server/channel\n", '|');
			escape_print(win, " |^F| scroll-back in what was written by others\n", '|');
			escape_print(win, " |^B| scroll-back in what you wrote. enter to select a line\n", '|');
			escape_print(win, " |^V| Enter 1 character: used when entering ascii values < 32\n", '|');
			escape_print(win, " |TAB| auto-completion of commands and nick names\n", '|');
			escape_print(win, " |^G| close a channel (F9: undo close)\n", '|');
			escape_print(win, " |^C| terminate the program\n", '|');
			if (vc_list_data_only)
				escape_print(win, "*|^Y| toggle \"only show channels with new messages\"", '|');
			else
				escape_print(win, " |^Y| toggle \"only show channels with new messages\"", '|');
		}
		else if (screen_nr == 2)
		{
			escape_print(win, "\n", '|');
			escape_print(win, " |F1| this help\n", '|');
			escape_print(win, " |F2| store current configuration on disk\n", '|');

			if (n_servers == 1)
				escape_print(win, " |F3| \0032add server\003\n", '|');
			else
				escape_print(win, " |F3| add server\n", '|');

			escape_print(win, " |F4| switch to edit-line (shortcut for ^N)\n", '|');
			escape_print(win, " |F5|/|^L| redraw screen\n", '|');
			escape_print(win, " |F6| search in all windows for text\n", '|');
			escape_print(win, " |F7| close all channels with only \"NOTICE\" messages\n", '|');
			escape_print(win, " |F8| edit configuration\n", '|');
			escape_print(win, " |F9| undo last channel close\n", '|');
			escape_print(win, " |F10|/|^N| toggle between channel-list, edit-line and word-cloud\n", '|');
			escape_print(win, " |@/|... sends /... to other end\n", '|');
			escape_print(win, " |@@|... sends @... to other end\n", '|');
			escape_print(win, " |@|... goes to the first channel with \"...\" in its name |^J| for the next\n", '|');
		}
		else if (screen_nr == 3)
		{
			const char pairs[] = " Defined pairs: ";
			int x = sizeof pairs, index = 0, n_cpairs = get_n_cpairs();
			char *nc_vars = NULL;
			wchar_t block = 0;
			char *pblock = "\xe2\x96\x85";
			char *res = NULL;

			mbsrtowcs(&block, (const char **)&pblock, 1, NULL);

			escape_print(win, "\n If you're on IRCNet, OFTC or FreeNet, say hi to |flok|/|flok99|/|flok42|!\n\n", '|');

			waddstr(win -> win, " Compiled on " __DATE__ " " __TIME__ "\n\n");

			asprintf(&res, " Columns/rows: %dx%d\n\n", max_x, max_y);
			waddstr(win -> win, res);
			free(res);

			asprintf(&nc_vars, " Colors: %d, pairs: %d, pairs defined: %d, can change colors: %s,\n nick color pairs: %d\n", COLORS, COLOR_PAIRS, n_cpairs, can_change_color() ? "yes" : "no", n_nick_pairs);
			waddstr(win -> win, nc_vars);
			waddstr(win -> win, "\n");
			waddstr(win -> win, pairs);

			for(index=0; index<n_cpairs; index++)
			{
				color_on(win, index);

				waddnwstr(win -> win, &block, 1);

				color_off(win, index);

				if (++x == help_win_width - 2)
				{
					waddstr(win -> win, "\n ");

					x = 1;
				}
			}

			free(nc_vars);
		}

		mydoupdate();

		c = wait_for_keypress(FALSE);

		if (c == 'q' || c == 'Q' || c == -1)
			break;
		else if (c == 3)
			exit_fi();
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
		{
			if (screen_nr == 0)
				break;

			screen_nr--;
		}
		else if (c == KEY_RIGHT)
		{
			if (screen_nr < 3)
				screen_nr++;
			else
				wrong_key();
		}
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();
}

void add_server_menu(void)
{
	const char *nick = NULL, *s = NULL;

	s = edit_box(60, "^Add server: enter hostname:port^", NULL);

	if (s)
		nick = edit_box(60, "^Add server: enter nick^", NULL);

	if (nick)
	{
		add_server(s, nick, "notset", nick, nick, s);

		popup_notify(FALSE, "Server %s was added", s);
	}

	myfree(s);
	myfree(nick);
}

void user_channel_menu(void)
{
	popup_notify(FALSE, "No special functions for 1-on-1 channels yet");
}

/* use_getch: used when no valid configuration was loaded */
yna_reply_t yesno_box(BOOL use_getch, const char *title, const char *q, BOOL allow_abort)
{
	NEWWIN *bwin = NULL, *win = NULL;
	int q_len = max(strlen(title), strlen(q));
	yna_reply_t rc = NO;

	create_win_border(q_len + 5, 4, title, &bwin, &win);

	for(;;)
	{
		int c = -1;

		werase(win -> win);

		escape_print_xy(win, 1, 1, q);

		if (allow_abort)
			escape_print_xy(win, 2, 1, "^y^es / ^n^o / ^a^bort");
		else
			escape_print_xy(win, 2, 1, "^y^es / ^n^o");

		mydoupdate();

		if (use_getch)
			c = getch();
		else
			c = wait_for_keypress(FALSE);

		if (c == 3 || c == 'n' || c == 'N')
		{
			rc = NO;
			break;
		}

		if (c == 'y' || c == 'Y')
		{
			rc = YES;
			break;
		}

		if ((c == 7 || c == 'a' || c == 'A') && allow_abort)
		{
			rc = ABORT;
			break;
		}

		wrong_key();
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	return rc;
}

void put_field(NEWWIN *win, const char *what, int y, int x, short fg, short bg)
{
	short dummy = get_color_ncurses(fg, bg);

	color_on(win, dummy);

	mvwprintw(win -> win, y, x, what);

	color_off(win, dummy);
}

void put_colors(NEWWIN *win, int y, int x)
{
	put_field(win, "BLACK  ", y + 0, x, COLOR_WHITE, COLOR_BLACK);
	put_field(win, "RED    ", y + 1, x, COLOR_BLACK, COLOR_RED);
	put_field(win, "GREEN  ", y + 2, x, COLOR_BLACK, COLOR_GREEN);
	put_field(win, "YELLOW ", y + 3, x, COLOR_BLACK, COLOR_YELLOW);
	put_field(win, "BLUE   ", y + 4, x, COLOR_BLACK, COLOR_BLUE);
	put_field(win, "MAGENTA", y + 5, x, COLOR_BLACK, COLOR_MAGENTA);
	put_field(win, "CYAN   ", y + 6, x, COLOR_BLACK, COLOR_CYAN);
	put_field(win, "WHITE  ", y + 7, x, COLOR_BLACK, COLOR_WHITE);
	put_field(win, "default", y + 8, x, -1, -1);
}

void put_cursor(NEWWIN *win, int y, int x, BOOL selected)
{
	mywattron(win -> win, A_REVERSE);

	if (selected)
		mvwprintw(win -> win, y, x, ">");
	else
		mvwprintw(win -> win, y, x, " ");

	mywattroff(win -> win, A_REVERSE);
}

int choose_colorpair(const int default_pair)
{
	int color = default_pair;
	NEWWIN *bwin = NULL, *win = NULL;
	BOOL ret_ok = TRUE, update_buttons = TRUE;
	short fg = 0, bg = 0;
	int column = 0;

	pair_content(default_pair, &fg, &bg);

	if (fg == -1)
		fg = 8;
	if (bg == -1)
		bg = 8;

	create_win_border(50, 15, " Navigate with cursor keys, tab+enter to exit ", &bwin, &win);

	werase(win -> win);

	mywattron(win -> win, A_UNDERLINE);
	mvwprintw(win -> win, 1, 10, "foreground");
	mvwprintw(win -> win, 1, 26, "background");
	mywattroff(win -> win, A_UNDERLINE);

	put_colors(win, 2, 10);
	put_colors(win, 2, 26);

	for(;;)
	{
		int c = -1, old_fg = fg, old_bg = bg;

		put_cursor(win, 2 + fg, 8, column == 0);
		put_cursor(win, 2 + bg, 24, column == 1);

		if (update_buttons)
		{
			if (ret_ok == TRUE)
				escape_print_xy(win, 13, 1, "^[ OK ]^ [ CANCEL ]");
			else
				escape_print_xy(win, 13, 1, "[ OK ] ^[ CANCEL ]^");

			update_buttons = FALSE;
		}

		c = wait_for_keypress(FALSE);

		if (c == 9)
		{
			ret_ok = !ret_ok;
			update_buttons = TRUE;
		}
		else if (c == 3)
		{
			color = default_pair;
			break;
		}
		else if (c == 13)
		{
			if (!ret_ok)
			{
				color = default_pair;
				break;
			}

			if (fg == 8)
				fg = -1;
			if (bg == 8)
				bg = -1;

			color = get_color_ncurses(fg, bg);

			if (fg == bg && fg != -1)
			{
				if (yesno_box(FALSE, "Use selected colors?", "Foreground and background colors are the same: are you sure?", FALSE) == YES)
					break;
			}
			else
			{
				break;
			}
		}
		else if (c == KEY_LEFT)
		{
			if (column > 0)
				column--;
			else
				wrong_key();
		}
		else if (c == KEY_RIGHT)
		{
			if (column < 1)
				column++;
			else
				wrong_key();
		}
		else if (c == KEY_UP)
		{
			if (column == 0)
			{
				if (fg)
					fg--;
				else
					wrong_key();
			}
			else
			{
				if (bg)
					bg--;
				else
					wrong_key();
			}
		}
		else if (c == KEY_DOWN)
		{
			if (column == 0)
			{
				if (fg < 8)
					fg++;
				else
					wrong_key();
			}
			else
			{
				if (bg < 8)
					bg++;
				else
					wrong_key();
			}
		}
		else
		{
			wrong_key();
		}

		if (fg != old_fg)
			mvwprintw(win -> win, 2 + old_fg, 8, " ");

		if (bg != old_bg)
			mvwprintw(win -> win, 2 + old_bg, 24, " ");
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	return color;
}

BOOL configure_firc(void)
{
	BOOL force_redraw = TRUE;
	NEWWIN *bwin = NULL, *win = NULL;
	int cursor = 0, offset = 0;
	BOOL redraw = TRUE;
	const int win_height = 20, win_width = 68, content_height = win_height - 5, content_offset = 4;
	int cnt = 0, cnf_len = 0, val_pos = 50;
	const char title[] = " Configure f-irc, key right to select, key left to exit ";

	for(cnt=0; cnf_pars[cnt].p != NULL; cnt++)
	{
		int l = strlen(cnf_pars[cnt].name);

		if (l > cnf_len)
			cnf_len = l;
	}

	if (cnf_len + 2 < val_pos)
		val_pos = cnf_len + 2;

	create_win_border(win_width, win_height, title, &bwin, &win);

	for(;;)
	{
		int prev_pos = offset + cursor;
		int c = -1;

		if (redraw)
		{
			int loop = 0, max_rows = cnt - offset;

			werase(win -> win);

			escape_print_xy(win, 1, 1, "^F2^: store current configuration on disk");
			escape_print_xy(win, 2, 1, "^s^: edit server(s)  ^l^ edit favorites-list");

			for(loop=0; loop<min(max_rows, content_height); loop++)
			{
				int pos = offset + loop, sp = 0;
				char *str_out = (char *)calloc(1, win_width - 2 + 1);
				BOOL *pb = NULL;
				int *pi = NULL;
				char *pc = NULL;
				short cfg = -1, cbg = -1;

				snprintf(&str_out[0], val_pos - 2, "%s", cnf_pars[pos].name);

				switch(cnf_pars[pos].type)
				{
					case CNF_BOOL:
						pb = (BOOL *)cnf_pars[pos].p;

						if (*pb)
							sprintf(&str_out[val_pos], "ON");
						else
							sprintf(&str_out[val_pos], "OFF");
						break;

					case CNF_VALUE:
						pi = (int *)cnf_pars[pos].p;

						sprintf(&str_out[val_pos], "%d", *pi);
						break;

					case CNF_STRING:
						pc = *(char **)cnf_pars[pos].p;

						snprintf(&str_out[val_pos], (win_width - 2) - val_pos, "%s", str_or_nothing(pc));
						break;

					case CNF_COLOR:
						pi = (int *)cnf_pars[pos].p;

						pair_content(*pi, &cfg, &cbg);

						sprintf(&str_out[val_pos], "%s,%s", color_to_str(cfg), color_to_str(cbg));
						break;
				}

				str_out[win_width - 2] = 0x00;

				for(sp=0; sp<win_width - 2; sp++)
				{
					if (str_out[sp] == 0x00)
						str_out[sp] = ' ';
				}

				if (loop == cursor)
					mywattron(win -> win, A_REVERSE);

				mvwprintw(win -> win, content_offset + loop, 1, "%s", str_out);

				if (loop == cursor)
					mywattroff(win -> win, A_REVERSE);

				free(str_out);
			}

			mydoupdate();

			redraw = FALSE;
		}

		c = wait_for_keypress(FALSE);

		if (c == KEY_LEFT || c == 3 || toupper(c) == 'Q')
			break;
		else if (c == KEY_LEFT || (c == KEY_MOUSE && right_mouse_button_clicked()))
			break;

		if (c == KEY_F(2))
		{
                        char *err_msg = NULL;

                        (void)save_config(TRUE, &err_msg);
                        popup_notify(FALSE, "%s", err_msg);

                        free(err_msg);

		}
		else if (c == 's')
		{
			for(;;)
			{
				int sr = select_server();

				if (sr == -1)
					break;

				edit_server(sr);
			}
		}
		else if (c == 'l')
		{
                        char *err_msg = NULL;

			edit_favorites();

                        if (save_config(TRUE, &err_msg) == FALSE)
			{
				popup_notify(FALSE, "%s", err_msg);

				free(err_msg);
			}
		}
		else if (c == KEY_RIGHT || c == 13 || c == ' ')
		{
			int pos = offset + cursor;
			BOOL *pb = NULL;
			int *pi = NULL;
			char *pi_str = NULL;
			char *pc = NULL;
			const char *new_str = NULL;
			char *q = NULL;

			asprintf(&q, "%s:", cnf_pars[pos].name);

			q[0] = toupper(q[0]);

			switch(cnf_pars[pos].type)
			{
				case CNF_BOOL:
					pb = (BOOL *)cnf_pars[pos].p;
					*pb = onoff_box(q, *pb);
					break;

				case CNF_VALUE:
					pi = (int *)cnf_pars[pos].p;
					asprintf(&pi_str, "%d", *pi);

					new_str = edit_box(val_pos, q, pi_str);

					if (new_str)
					{
						*pi= atoi(new_str);
						myfree(new_str);
					}

					free(pi_str);
					break;

				case CNF_STRING:
					pc = *(char **)cnf_pars[pos].p;

					new_str = edit_box(val_pos, q, pc);

					if (new_str)
					{
						free(pc);

						*(char **)cnf_pars[pos].p = (char *)new_str;
					}

					break;

				case CNF_COLOR:
					pi = (int *)cnf_pars[pos].p;
					*pi = choose_colorpair(*pi);
					break;
			}

			if (cnf_pars[pos].dofunc)
				(cnf_pars[pos].dofunc)();

			free(q);

			redraw = TRUE;
		}
		else if (c == KEY_UP)
		{
			if (cursor > 0)
				cursor--;
			else if (offset > 0)
				offset--;
			else
				wrong_key();
		}
		else if (c == KEY_DOWN)
		{
			if (cursor < content_height - 1 && cursor + offset < cnt - 1)
				cursor++;
			else if (cursor + offset < cnt - 1)
				offset++;
			else
				wrong_key();
		}
		else if (c == KEY_HOME)
		{
			cursor = offset = 0;
		}
		else if (c == KEY_END)
		{
			cursor = content_height - 1;
			offset = cnt - cursor - 1;
		}
		else
		{
			wrong_key();
		}

		if (cursor + offset != prev_pos)
			redraw = TRUE;
	}

	delete_window(win);
	delete_window(bwin);

	mydoupdate();

	return force_redraw;
}

void close_notice_channels(void)
{
	if (yesno_box(FALSE, "Close all \"NOTICE\" channels", "Are you sure?", FALSE) == YES)
	{
		int sr = 0;
		NEWWIN *win = NULL, *bwin = NULL;

		create_win_border(30, 3, "Please wait", &bwin, &win);
		mvwprintw(win -> win, 1, 1, "Closing \"NOTICE\" channels...");
		mydoupdate();

		for(sr = 0; sr < n_servers; sr++)
		{
			int ch = 0;
			server *ps = &server_list[sr];
			BOOL left = FALSE;

			do
			{
				left = FALSE;

				for(ch=1; ch<ps -> n_channels; ch++)
				{
					channel *pc = &ps -> pchannels[ch];

					if (is_channel(pc -> channel_name))
						continue;

					if (pc -> recvd_non_notice == FALSE)
					{
						cmd_LEAVE(sr, ch, NULL);
						left = TRUE;
						break;
					}
				}
			}
			while(left);
		}

		delete_window(win);
		delete_window(bwin);

		mydoupdate();
	}
}

void show_channel_history(void)
{
	NEWWIN *bwin = NULL, *win = NULL;
	create_win_border(max_x - (theme.channellist_window_width + 2), max_y - 4, "left cursor key: exit, / to search", &bwin, &win);

	scrollback(win, cur_channel() -> pbuffer, cur_server() -> nickname, FALSE, FALSE);

	delete_window(win);
	delete_window(bwin);
}

void select_own_history_line(void)
{
	char *selected_line = NULL;
	int sr = -1, ch = -1;
	NEWWIN *bwin = NULL, *win = NULL;

	create_win_border(max_x - (theme.channellist_window_width + 2), max_y - 4, "right cursor key: select, left: cancel", &bwin, &win);

	if (scrollback_and_select(win, cur_channel() -> input_buffer, &selected_line, &sr, &ch, cur_server() -> nickname, FALSE, FALSE))
	{
		if (utf8_strlen(cur_channel() -> input) > 0)
		{
			char *temp = utf8_get_utf8(cur_channel() -> input);

			add_to_buffer(cur_channel() -> input_buffer, temp, cur_server() -> nickname, FALSE, current_server, current_server_channel_nr);

			myfree(temp);
		}

		utf8_truncate(cur_channel() -> input, 0);
		utf8_strcat_ascii(cur_channel() -> input, selected_line);

		myfree(selected_line);
	}

	delete_window(win);
	delete_window(bwin);
}
